"""Test yaml test cases formatters."""

import typing
import pathlib

import pytest
from pglast import parser

from tests import TEST_FILE, conftest
from pgrubic import core
from pgrubic.core import errors


@pytest.mark.parametrize(
    ("test_formatter", "test_id", "test_case"),
    conftest.load_test_cases(
        test_case_type=conftest.TestCaseType.FORMATTER,
        directory=pathlib.Path("tests/fixtures/formatters"),
    ),
)
def test_formatters(
    formatter: core.Formatter,
    test_formatter: str,
    test_id: str,
    test_case: dict[str, str],
) -> None:
    """Test formatters."""
    config_overrides: dict[str, typing.Any] = typing.cast(
        dict[str, typing.Any],
        test_case.get("config", {}),
    )

    # Apply overrides to global configuration
    conftest.update_config(formatter.config, config_overrides)

    result = formatter.format(
        source_file=TEST_FILE,
        source_code=test_case["sql"],
    )

    assert (
        result.formatted_source_code == test_case["expected"]
    ), f"Test failed for formatter: `{test_formatter}` in `{test_id}`"

    # Check that the formatted source code is valid
    try:
        parser.parse_sql(result.formatted_source_code)
    except parser.ParseError as error:
        msg = f"Formatted code is not a valid syntax: {error!s}"
        raise ValueError(msg) from error


def test_format_parse_error(formatter: core.Formatter) -> None:
    """Test format."""
    source_code = "SELECT * FROM;"
    with pytest.raises(errors.ParseError):
        formatter.format(source_file=TEST_FILE, source_code=source_code)


def test_new_line_before_semicolon(formatter: core.Formatter) -> None:
    """Test new line before semicolon."""
    source_code = "select 1;"
    expected_output: str = "SELECT 1\n;\n"

    formatter.config.format.new_line_before_semicolon = True

    result = formatter.format(
        source_file=TEST_FILE,
        source_code=source_code,
    )

    assert result.formatted_source_code == expected_output



"""Checker for adding of stored generated column."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class AddingStoredGeneratedColumn(linter.BaseChecker):
    """## **What it does**
    Checks adding of stored generated column.

    ## **Why not?**
    Adding an stored generated column to an already populated table will
    have to backfill the newly added column, causing the table to be locked
    in which no other operations can be performed on the table for the duration
    of the backfill. This will cause downtime if the table is concurrently
    being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    A trigger might be a safer option.
    """

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if (
            ancestors.find_nearest(ast.AlterTableCmd)
            and node.contype == enums.ConstrType.CONSTR_GENERATED
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Adding stored generated column is not safe",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="A trigger might be a safer option",
                ),
            )



"""Checker for implicit constraint name."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class ImplicitConstraintName(linter.BaseChecker):
    """## **What it does**
    Checks that constraint is explicitly named and not relying on implicit naming.

    ## **Why not?**
    Explicitly naming a constraint is a good practice. It improves your code clarity and
    makes it more readable.

    A good naming convention makes your code easier to read and understand.

    ## **When should you?**
    Never almost.

    ## **Use instead:**
    Name your constraint according to the set naming convention.
    """

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if not node.conname and node.contype in (
            enums.ConstrType.CONSTR_CHECK,
            enums.ConstrType.CONSTR_PRIMARY,
            enums.ConstrType.CONSTR_UNIQUE,
            enums.ConstrType.CONSTR_EXCLUSION,
            enums.ConstrType.CONSTR_FOREIGN,
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer named constraint",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Name your constraint explicitly",
                ),
            )



"""Test noqa."""

import typing
import pathlib

import pytest
from colorama import Fore, Style

from tests import TEST_FILE
from pgrubic.core import noqa, errors


def test_extract_star_ignore_from_inline_comments() -> None:
    """Test extract star ignore from inline comments."""
    source_code: str = """
    -- noqa
    CREATE TABLE tbl (activated date);
    """

    inline_ignores: list[noqa.NoQaDirective] = noqa.extract_ignores(
        source_file=TEST_FILE,
        source_code=source_code,
    )

    assert inline_ignores[0].rule == noqa.A_STAR


def test_extract_ignores() -> None:
    """Test extract ignores from inline comments."""
    source_code: str = """-- pgrubic: noqa: NM016, GN001
    CREATE TABLE tbl (activated date);
    """

    inline_ignores: list[noqa.NoQaDirective] = noqa.extract_ignores(
        source_file=str(TEST_FILE),
        source_code=source_code,
    )

    assert inline_ignores[0].rule == "NM016"
    assert inline_ignores[1].rule == "GN001"


def test_extract_ignores_length() -> None:
    """Test extract ignore from inline comments length."""
    source_code: str = """
    -- noqa: NM016, GN001
    CREATE TABLE tbl (activated date);
    """

    inline_ignores: list[noqa.NoQaDirective] = noqa.extract_ignores(
        source_file=TEST_FILE,
        source_code=source_code,
    )

    expected_ignores_length: int = 2

    assert len(inline_ignores) == expected_ignores_length


def test_wrongly_formed_inline_ignores_from_inline_comments(capfd: typing.Any) -> None:
    """Test extract ignore from inline comments."""
    source_code: str = """
    -- noqa NM016, GN001
    CREATE TABLE tbl (activated date);
    """

    noqa.extract_ignores(source_file=TEST_FILE, source_code=source_code)

    _, err = capfd.readouterr()

    assert (
        err
        == f"{Fore.YELLOW}Warning: Malformed `noqa` directive at location 5. Expected `noqa: <rules>`{Style.RESET_ALL}\n"  # noqa: E501
    )


def test_report_specific_unused_ignores(
    capfd: typing.Any,
) -> None:
    """Test report specific unused ignores."""
    source_code: str = """
    -- noqa: NM016
    CREATE TABLE tbl (activated date);
    """

    inline_ignores: list[noqa.NoQaDirective] = noqa.extract_ignores(
        source_file=TEST_FILE,
        source_code=source_code,
    )

    noqa.report_unused_ignores(source_file=TEST_FILE, inline_ignores=inline_ignores)
    out, _ = capfd.readouterr()
    assert (
        out
        == f"{TEST_FILE}:3:52: {Fore.YELLOW}Unused noqa directive{Style.RESET_ALL} (unused: {Fore.RED}{Style.BRIGHT}NM016{Style.RESET_ALL})\n"  # noqa: E501
    )


def test_missing_statement_terminator() -> None:
    """Test missing statement terminator."""
    source_code: str = "SELECT * FROM tab"

    with pytest.raises(errors.MissingStatementTerminatorError):
        noqa.extract_comments(
            source_file=TEST_FILE,
            source_code=source_code,
        )


def test_report_general_unused_ignores(
    capfd: typing.Any,
) -> None:
    """Test report general unused ignores."""
    source_code: str = """
    -- noqa
    CREATE TABLE tbl (activated date);
    """

    inline_ignores: list[noqa.NoQaDirective] = noqa.extract_ignores(
        source_file=TEST_FILE,
        source_code=source_code,
    )

    noqa.report_unused_ignores(source_file=TEST_FILE, inline_ignores=inline_ignores)
    out, _ = capfd.readouterr()
    assert (
        out
        == f"{TEST_FILE}:3:45: {Fore.YELLOW}Unused noqa directive{Style.RESET_ALL} (unused: {Fore.RED}{Style.BRIGHT}{noqa.A_STAR}{Style.RESET_ALL})\n"  # noqa: E501
    )


def test_add_file_level_general_ignore(tmp_path: pathlib.Path) -> None:
    """Test add file level general ignore."""
    directory = tmp_path / "sub"
    directory.mkdir()

    source_file1 = directory / "source_file1.sql"
    source_file1.write_text("SELECT * FROM tab")

    source_file2 = directory / "source_file2.sql"
    source_file2.write_text("-- pgrubic: noqa\n SELECT * FROM tab")

    modified_sources = noqa.add_file_level_general_ignore(
        sources={source_file1, source_file2},
    )

    assert modified_sources == 1



"""Formatter for view."""

from pglast import ast, enums, stream, printers


@printers.node_printer(ast.ViewStmt, override=True)
def view_stmt(node: ast.ViewStmt, output: stream.RawStream) -> None:
    """Printer for ViewStmt."""
    output.write("CREATE")
    output.space()
    if node.replace:
        output.write("OR REPLACE")
        output.space()

    if node.view.relpersistence == enums.RELPERSISTENCE_TEMP:
        output.write("TEMPORARY")
        output.space()
    elif node.view.relpersistence == enums.RELPERSISTENCE_UNLOGGED:
        output.write("UNLOGGED")
        output.space()

    output.write("VIEW")
    output.space()
    output.print_node(node.view)

    if node.aliases:
        output.space()
        with output.expression(need_parens=True):
            output.print_list(node.aliases, are_names=True)
    output.space()

    if node.options:  # pragma: no cover
        output.write("WITH")
        output.space()
        with output.expression(need_parens=True):
            output.print_list(node.options)
        output.newline()
        output.space(2)

    output.write("AS")
    output.newline()
    with output.push_indent():
        output.print_node(node.query)

    if node.withCheckOption:
        output.newline()
        output.space()
        printers.ddl.view_check_option_printer(node.withCheckOption, node, output)



"""Checker for wrongly typed required columns."""

from pglast import ast, visitors
from pglast.printers import dml

from pgrubic import get_fully_qualified_name
from pgrubic.core import config, linter


class WronglyTypedRequiredColumn(linter.BaseChecker):
    """## **What it does**
    Checks for wrongly typed required columns.

    ## **Why not?**
    If a column has been specified as required and you have typed it wrongly,
    you are probably doing something wrong.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Right data type for the required column.

    ## **Configuration**
    `required-columns`: List of required columns along with their data types.
    """

    is_auto_fixable = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        for column in self.config.lint.required_columns:
            if column.name == node.colname:
                fully_qualified_type_name = get_fully_qualified_name(
                    node.typeName.names,
                )

                prettified_type = fully_qualified_type_name

                if fully_qualified_type_name in dml.system_types:
                    prettified_type = dml.system_types[fully_qualified_type_name][0]

                if column.data_type != prettified_type:
                    self.violations.add(
                        linter.Violation(
                            rule_code=self.code,
                            rule_name=self.name,
                            rule_category=self.category,
                            line_number=self.line_number,
                            column_offset=self.column_offset,
                            line=self.line,
                            statement_location=self.statement_location,
                            description=f"Column '{node.colname}' expected type is"
                            f" '{column.data_type}', found"
                            f" '{prettified_type}'",
                            is_auto_fixable=self.is_auto_fixable,
                            is_fix_enabled=self.is_fix_enabled,
                            help="Use the right data type for the required column",
                        ),
                    )

                    self._fix(node, column)

    def _fix(self, node: ast.ColumnDef, column: config.Column) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": column.data_type,
                },
            ),
        )



"""Linter."""

from __future__ import annotations

import sys
import copy
import typing
import fnmatch
import functools

from pglast import ast, parser, stream, visitors
from colorama import Fore, Style
from caseconverter import kebabcase

from pgrubic import DOCUMENTATION_URL
from pgrubic.core import noqa, config, errors, formatter

if typing.TYPE_CHECKING:
    from collections import abc  # pragma: no cover


class Violation(typing.NamedTuple):
    """Representation of rule violation."""

    rule_code: str
    rule_name: str
    rule_category: str
    line_number: int
    column_offset: int
    line: str
    statement_location: int
    description: str
    is_auto_fixable: bool
    is_fix_enabled: bool
    help: str | None = None


class LintResult(typing.NamedTuple):
    """Lint Result."""

    source_file: str
    violations: set[Violation]
    fixed_sql: str | None = None


class ViolationStats(typing.NamedTuple):
    """Violation Stats."""

    total: int
    auto_fixable: int
    fix_enabled: int


class CheckerMeta(type):
    """Metaclass for Checker. This metaclass handles both runtime and subclass method
    decorations. It is originally created for method decorations but could be extended in
    the future.
    """

    def __new__(
        cls: type[CheckerMeta],
        name: str,
        bases: tuple[typing.Any],
        dct: dict[str, typing.Any],
    ) -> typing.Any:
        """Add method decorations."""
        for attr_name, attr_value in dct.items():
            if callable(attr_value) and attr_name.startswith("_fix"):
                dct[attr_name] = cls._apply_fix(attr_value)

            if callable(attr_value) and attr_name.startswith("visit_"):
                dct[attr_name] = cls._set_locations(attr_value)

        return super().__new__(cls, name, bases, dct)

    @staticmethod
    def _set_locations(
        func: abc.Callable[..., typing.Any],
    ) -> abc.Callable[..., typing.Any]:
        """Helper method to set locations for node."""

        @functools.wraps(func)
        def wrapper(
            checker: BaseChecker,
            ancestors: visitors.Ancestor,
            node: ast.Node,
        ) -> typing.Any:
            # some nodes have location attribute which is different from node location
            # for example ast.CreateTablespaceStmt while some nodes do not carry
            # location at all
            if hasattr(node, "location") and isinstance(node.location, int):
                checker.node_location = checker.statement_location + node.location
            else:
                checker.node_location = checker.statement_location + len(
                    checker.statement,
                )

            # get the position of the newline just before our node location,
            line_start = (
                checker.source_code.rfind(noqa.NEW_LINE, 0, checker.node_location) + 1
            )
            # get the position of the newline just after our node location
            line_end = checker.source_code.find(noqa.NEW_LINE, checker.node_location)

            # line number is number of newlines before our node location,
            # increment by 1 to land on the actual node
            checker.line_number = (
                checker.source_code[: checker.node_location].count(noqa.NEW_LINE) + 1
            )
            checker.column_offset = checker.node_location - line_start + 1

            # If a node has no location, we return the whole statement instead
            if hasattr(node, "location") and isinstance(node.location, int):
                checker.line = checker.source_code[line_start:line_end]
            else:
                checker.line = checker.statement.strip()

            return func(checker, ancestors, node)

        return wrapper

    @staticmethod
    def _apply_fix(
        func: abc.Callable[..., typing.Any],
    ) -> abc.Callable[..., typing.Any]:
        """Helper method to apply fix only if it is applicable."""

        @functools.wraps(func)
        def wrapper(
            checker: BaseChecker,
            *args: typing.Any,
            **kwargs: typing.Any,
        ) -> typing.Any:
            if not checker.config.lint.fix:
                return None

            if not checker.is_fix_applicable:
                return None

            return func(checker, *args, **kwargs)

        return wrapper


class BaseChecker(visitors.Visitor, metaclass=CheckerMeta):  # type: ignore[misc]
    """Define a lint rule, and store all the nodes that violate it."""

    # Should not be set directly
    # as they are set in __init_subclass__
    code: str
    name: str
    category: str

    # Is this rule automatically fixable?
    is_auto_fixable: bool = False

    # Attributes shared among all subclasses
    config: config.Config
    inline_ignores: list[noqa.NoQaDirective]
    source_file: str
    source_code: str

    statement_location: int
    node_location: int
    line_number: int
    column_offset: int
    statement: str
    line: str

    def __init__(self) -> None:
        """Initialize variables."""
        self.violations: set[Violation] = set()

    def __init_subclass__(cls, **kwargs: typing.Any) -> None:
        """Set code, name and category attributes for subclasses."""
        cls.code = cls.__module__.split(".")[-1]
        cls.name = kebabcase(cls.__name__)
        cls.category = cls.__module__.split(".")[-2]

    def visit(self, node: ast.Node, ancestors: visitors.Ancestor) -> None:
        """Visit the node."""

    @property
    def is_fix_enabled(self) -> bool:
        """Check if fix is enabled according to config."""
        # if a rule is not auto fixable, there is no need to check
        # if its fix is enabled, in which case, we return False
        if not self.is_auto_fixable:
            return False

        return not (
            (
                self.config.lint.fixable
                and not any(
                    fnmatch.fnmatch(self.code, pattern + "*")
                    for pattern in self.config.lint.fixable
                )
            )
            or any(
                fnmatch.fnmatch(self.code, pattern + "*")
                for pattern in self.config.lint.unfixable
            )
        )

    @property
    def is_fix_applicable(self) -> bool:
        """Check if fix can be applied."""
        if not self.is_auto_fixable:
            return False

        if not self.is_fix_enabled:
            return False

        # if the violation has been suppressed by noqa, there is no need to try to fix it
        for inline_ignore in self.inline_ignores:
            if (
                (
                    self.statement_location == inline_ignore.location
                    or self.source_file == inline_ignore.source_file
                )
                and inline_ignore.rule in (noqa.A_STAR, self.code)
                and not self.config.lint.ignore_noqa
            ):
                return False

        return True


class Linter:
    """Holds all lint rules, and runs them against a source code."""

    def __init__(
        self,
        config: config.Config,
        formatters: typing.Callable[
            [],
            set[typing.Callable[[], None]],
        ],
    ) -> None:
        """Initialize variables."""
        self.checkers: set[BaseChecker] = set()
        self.config = config
        self.formatter = formatter.Formatter(
            config=config,
            formatters=formatters,
        )

    @staticmethod
    def _skip_suppressed_violations(
        *,
        source_file: str,
        checker: BaseChecker,
        inline_ignores: list[noqa.NoQaDirective],
    ) -> None:
        """Skip suppressed violations.

        Parameters:
        ----------
        source_file: str
            Path to the source file.
        checker: BaseChecker
            Lint rule checker.
        inline_ignores: list[NoQaDirective]
            Inline noqa directives.

        Returns:
        -------
        None
        """
        for inline_ignore in inline_ignores:
            suppressed_violations: set[Violation] = {
                violation
                for violation in checker.violations
                if (
                    (
                        violation.statement_location == inline_ignore.location
                        or source_file == inline_ignore.source_file
                    )
                    and (inline_ignore.rule in (noqa.A_STAR, checker.code))
                )
            }

            if suppressed_violations:
                inline_ignore.used = True

                checker.violations = {
                    violation
                    for violation in checker.violations
                    if violation not in suppressed_violations
                }

    @staticmethod
    def get_violation_stats(violations: set[Violation]) -> ViolationStats:
        """Get violation stats.

        Parameters:
        ----------
        violations: set[Violation]
            Violations to get stats from.

        Returns:
        -------
        ViolationStats
            Violation stats.
        """
        return ViolationStats(
            total=len(violations),
            auto_fixable=sum(1 for violation in violations if violation.is_auto_fixable),
            fix_enabled=sum(1 for violation in violations if violation.is_fix_enabled),
        )

    @staticmethod
    def print_violations(
        *,
        violations: set[Violation],
        source_file: str,
    ) -> None:
        """Print all violations collected by a checker.

        Parameters:
        ----------
        violations: set[Violation]
            Violations to print.
        source_file: str
            Path to the source file.

        Returns:
        -------
        None
        """
        for violation in violations:
            # if not checker.is_fix_applicable:
            sys.stdout.write(
                f"{noqa.NEW_LINE}{source_file}:{violation.line_number}:{violation.column_offset}:"
                f" \033]8;;{DOCUMENTATION_URL}/rules/{violation.rule_category}/{violation.rule_name}{Style.RESET_ALL}\033\\{Fore.RED}{Style.BRIGHT}{violation.rule_code}{Style.RESET_ALL}\033]8;;\033\\:"  # noqa: E501
                f" {violation.description}{noqa.NEW_LINE}",
            )

            for idx, line in enumerate(
                violation.line.splitlines(keepends=False),
                start=violation.line_number - violation.line.count(noqa.NEW_LINE),
            ):
                sys.stdout.write(
                    f"{Fore.BLUE}{idx} | {Style.RESET_ALL}{Fore.RED}{Style.BRIGHT}{line}{Style.RESET_ALL}{noqa.NEW_LINE}",  # noqa: E501
                )
                # in order to have arrow pointing to the violation, we need to shift
                # the screen by the length of the line_number as well as 2 spaces
                # used above between the separator (|)
                (
                    sys.stdout.write(
                        " "
                        * (violation.column_offset + len(str(violation.line_number)) + 2)
                        + "^"
                        + noqa.NEW_LINE,
                    )
                    if idx == violation.line_number
                    else None
                )

    def run(self, *, source_file: str, source_code: str) -> LintResult:
        """Run rules on a source code.

        Parameters:
        ----------
        source_file: str
            Path to the source file.
        source_code: str
            Source code to lint.

        Returns:
        -------
        LintResult
            Lint result.

        Raises:
        ------
        ParseError
            If there is an error parsing source code.
        """
        fixed_statements: list[str] = []

        inline_ignores: list[noqa.NoQaDirective] = noqa.extract_ignores(
            source_file=source_file,
            source_code=source_code,
        )

        violations: set[Violation] = set()

        BaseChecker.inline_ignores = inline_ignores
        BaseChecker.source_code = source_code
        BaseChecker.source_file = source_file
        BaseChecker.config = self.config

        for statement in noqa.extract_statement_locations(
            source_file=source_file,
            source_code=source_code,
        ):
            try:
                original_tree: ast.Node = parser.parse_sql(statement.text)
                copied_tree: ast.Node = copy.deepcopy(original_tree)

                comments = noqa.extract_comments(
                    source_file=source_file,
                    source_code=statement.text,
                )

            except parser.ParseError as error:
                msg = "Error parsing source code: "
                raise errors.ParseError(f"{source_file}: " + msg + str(error)) from error

            BaseChecker.statement = statement.text
            BaseChecker.statement_location = statement.start_location

            for checker in self.checkers:
                checker.violations = set()

                checker(copied_tree)

                if not self.config.lint.ignore_noqa:
                    self._skip_suppressed_violations(
                        source_file=source_file,
                        checker=checker,
                        inline_ignores=inline_ignores,
                    )

                violations.update(checker.violations)

            if original_tree != copied_tree:
                fixed_statement = stream.IndentedStream(
                    comments=comments,
                    semicolon_after_last_statement=False,
                    special_functions=True,
                    separate_statements=self.config.format.lines_between_statements,
                    remove_pg_catalog_from_functions=self.config.format.remove_pg_catalog_from_functions,
                    comma_at_eoln=not (self.config.format.comma_at_beginning),
                )(copied_tree)

                if self.config.format.new_line_before_semicolon:
                    fixed_statement += noqa.NEW_LINE + noqa.SEMI_COLON
                else:
                    fixed_statement += noqa.SEMI_COLON

                fixed_statements.append(fixed_statement)

            else:
                fixed_statements.append(statement.text.strip())

        fixed_source_code = (
            noqa.NEW_LINE + (noqa.NEW_LINE * self.config.format.lines_between_statements)
        ).join(
            fixed_statements,
        ) + noqa.NEW_LINE

        fix = None

        if parser.parse_sql(fixed_source_code) != parser.parse_sql(source_code):
            fix = fixed_source_code

        noqa.report_unused_ignores(
            source_file=source_file,
            inline_ignores=inline_ignores,
        )

        return LintResult(source_file=source_file, violations=violations, fixed_sql=fix)



"""Test linter."""

import pytest

from pgrubic import core
from pgrubic.core import errors

SOURCE_FILE = "linter.sql"


def test_linter_violations(
    linter: core.Linter,
) -> None:
    """Test linter violations."""
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="SELECT a = NULL, 10 = b;",
    )

    expected_number_of_violations: int = 2

    assert len(linting_result.violations) == expected_number_of_violations


def test_linter_suppressed_all_violations(
    linter: core.Linter,
) -> None:
    """Test linter suppressed all violations."""
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="""
        -- noqa
        SELECT a = NULL, 10 = b;
        """,
    )

    assert len(linting_result.violations) == 0


def test_linter_suppressed_certain_violations(
    linter: core.Linter,
) -> None:
    """Test linter suppressed certain violations."""
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="""
        -- noqa: GN024
        SELECT a = NULL, 10 = b;
        """,
    )

    assert len(linting_result.violations) == 1


def test_linter_violations_fixed_sql(
    linter: core.Linter,
) -> None:
    """Test linter fixed sql."""
    linter.config.lint.fix = True
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="SELECT a = NULL;",
    )

    assert linting_result.fixed_sql == "SELECT a IS NULL;\n"


def test_linter_suppressed_violations_fixed_sql(
    linter: core.Linter,
) -> None:
    """Test linter suppressed violations fixed sql."""
    linter.config.lint.fix = True
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="""
        -- noqa: GN024
        SELECT a = NULL;
        """,
    )

    assert not linting_result.fixed_sql


def test_parse_error(linter: core.Linter) -> None:
    """Test parse error."""
    source_code: str = """
    CREATE TABLE tbl (activated);
    """

    with pytest.raises(errors.ParseError):
        linter.run(
            source_file=SOURCE_FILE,
            source_code=source_code,
        )


def test_new_line_before_semicolon(
    linter: core.Linter,
) -> None:
    """Test new line before semicolon."""
    linter.config.lint.fix = True
    linter.config.format.new_line_before_semicolon = True
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="SELECT a = NULL;",
    )

    assert linting_result.fixed_sql == "SELECT a IS NULL\n;\n"


def test_fix_enabledment(
    linter: core.Linter,
) -> None:
    """Test fix enabledment."""
    linter.config.lint.fix = True
    linter.config.lint.fixable = ["GN024"]
    linting_result = linter.run(
        source_file=SOURCE_FILE,
        source_code="SELECT a = NULL;",
    )

    assert linting_result.fixed_sql == "SELECT a IS NULL\n;\n"



"""Checker for vacuum full."""

from pglast import ast, visitors

from pgrubic.core import linter


class VacuumFull(linter.BaseChecker):
    """## **What it does**
    Checks vacuum full.

    ## **Why not?**
    When a table is being vacuumed with the **FULL** option, an **ACCESS EXCLUSIVE** lock
    is acquired on it, preventing any other database operations (both reads and writes)
    from operating on the table until the **VACUUM FULL** is finished.
    This will cause downtime if the table is concurrently being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    Have a look at **pg_repack** as an alternative.
    """

    def visit_VacuumStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.VacuumStmt,
    ) -> None:
        """Visit VacuumStmt."""
        options = [option.defname for option in node.options] if node.options else []

        if "full" in options:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Vacuum full found",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Have a look at pg_repack as an alternative",
                ),
            )



"""Checker for identifiers prefix with pg_."""

from pgrubic.core import linter
from pgrubic.rules.naming import CheckIdentifier


class PgPrefixIdentifier(CheckIdentifier):
    """## **What it does**
    Checks for identifiers prefix with **pg_**.

    ## **Why not?**
    From the documentation:

    Schema names beginning with **pg_** are reserved for system purposes and cannot be
    created by users.

    Since system table names begin with **pg_**, it is best to avoid such names to ensure
    that you won't suffer a conflict if some future version defines a system table named
    the same as your table. (With the default search path, an unqualified reference to
    your table name would then be resolved as the system table instead.)
    System tables will continue to follow the convention of having names beginning with
    **pg_**, so that they will not conflict with unqualified user-table names so long as
    users avoid the **pg_** prefix.

    Same thing applies to other objects such as functions, views, sequences etc.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Remove prefix **pg_** from identifier.
    """

    def _check_identifier(
        self,
        identifier: str,
        line_number: int,
        column_offset: int,
        line: str,
        statement_location: int,
    ) -> None:
        """Checks for identifiers prefix with pg_."""
        if identifier and identifier.strip().startswith("pg_"):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=line_number,
                    column_offset=column_offset,
                    line=line,
                    statement_location=statement_location,
                    description="Identifier should not use prefix `pg_`",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Remove prefix `pg_` from identifier",
                ),
            )



"""Checker for drop tablespace."""

from pglast import ast, visitors

from pgrubic.core import linter


class DropTablespace(linter.BaseChecker):
    """## **What it does**
    Checks dropping of tablespace.

    ## **Why not?**
    Not only that mistakenly dropping a tablespace can cause data loss, applications that
    rely on the data will break.

    ## **When should you?**
    If you really want to drop the tablespace.

    ## **Use instead:**
    No suggestions.
    """

    def visit_DropTableSpaceStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.DropTableSpaceStmt,
    ) -> None:
        """Visit DropTableSpaceStmt."""
        self.violations.add(
            linter.Violation(
                rule_code=self.code,
                rule_name=self.name,
                rule_category=self.category,
                line_number=self.line_number,
                column_offset=self.column_offset,
                line=self.line,
                statement_location=self.statement_location,
                description="Drop tablespace detected",
                is_auto_fixable=self.is_auto_fixable,
                is_fix_enabled=self.is_fix_enabled,
            ),
        )



"""Checker for timestamp without time zone."""

from pglast import ast, visitors

from pgrubic.core import linter


class TimestampWithoutTimezone(linter.BaseChecker):
    """## **What it does**
    Checks for usage of timestamp without time zone.

    ## **Why not?**
    timestamptz (also known as timestamp with time zone) zone records a single moment
    in time. Despite what the name says it doesn't store a timestamp, just a point
    in time described as the number of microseconds since January 1st, 2000 in UTC.
    You can insert values in any timezone and it'll store the point in time that value
    describes. By default it will display times in your current timezone, but you can
    use at time zone to display it in other time zones. Because it stores a point in
    time, it will do the right thing with arithmetic involving timestamps entered in
    different timezones - including between timestamps from the same statement_location on
    different sides of a daylight savings time change.

    timestamp (also known as timestamp without time zone) does not do any of that,
    it just stores a date and time you give it. You can think of it being a picture of
    a calendar and a clock rather than a point in time.
    Without additional information - the timezone - you don't know what time it records.
    Because of that, arithmetic between timestamps from different locations or between
    timestamps from summer and winter may give the wrong answer.

    So if what you want to store is a point in time, rather than a picture of a clock,
    use timestamptz (timestamp with time zone).

    ## **When should you?**
    If you're dealing with timestamps in an abstract way, or just saving and retrieving
    them from an app, where you aren't going to be doing arithmetic with them then
    timestamp might be suitable.

    ## **Use instead:**
    timestamptz (also known as timestamp with time zone).
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval == "timestamp":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer timestamp with timezone over"
                    " timestamp without timezone",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use timestamptz",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "timestamptz",
                },
            ),
        )



"""Checker for column data type change."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class ColumnDataTypeChange(linter.BaseChecker):
    """## **What it does**
    Checks column data type change.

    ## **Why not?**
    Changing a column's data type requires an **ACCESS EXCLUSIVE** lock on the table,
    preventing both reads and writes. Generally, such change forces a rewrite of the
    whole table and indexes with every other transactions being blocked for
    the duration of the rewrite.
    As an exception to rewrite, when changing the type of an existing column, if the USING
    clause does not change the column contents and the old type is either binary coercible
    to the new type or an unconstrained domain over the new type, a table rewrite is not
    needed. However, indexes must always be rebuilt unless the system can verify that the
    new index would be logically equivalent to the existing one.

    This change might also cause errors and break applications that rely on the
    column potentially disrupting business operations.

    ## **When should you?**
    If the above exception to table and index rewrite is applicable and the clients have
    been made aware of the new type.

    ## **Use instead:**
    1. Create a new column with the new type, nullable.
    2. Start writing data to the new column.
    3. Copy all data from the old column to the new column.
    4. Migrate clients to the new column.
    """

    def visit_AlterTableCmd(
        self,
        ancestors: visitors.Ancestor,
        node: ast.AlterTableCmd,
    ) -> None:
        """Visit AlterTableCmd."""
        if node.subtype == enums.AlterTableType.AT_AlterColumnType:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Column data type change is not safe",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use a new column with the new type",
                ),
            )



"""Checker for non concurrent refresh materialized view."""

from pglast import ast, visitors

from pgrubic.core import linter


class NonConcurrentRefreshMaterializedView(linter.BaseChecker):
    """## **What it does**
    Checks non-concurrent refresh materialized view.

    ## **Why not?**
    Refreshing a materialized view in non-concurrent mode will locks out reads on the
    materialized view until it is done. This will cause downtime if the materialized view
    is concurrently being read by other clients.

    ## **When should you?**
    If downtime is acceptable.

    ## **Use instead:**
    Refresh the materialized view in concurrent mode:
        **REFRESH MATERIALIZED VIEW CONCURRENTLY ..**

    Please note that **CONCURRENTLY** option is only allowed if there is at least one
    UNIQUE index on the materialized view which uses only column names and includes all
    rows; that is, it must not be an expression index or include a WHERE clause.
    """

    is_auto_fixable: bool = True

    def visit_RefreshMatViewStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.RefreshMatViewStmt,
    ) -> None:
        """Visit RefreshMatViewStmt."""
        if not node.concurrent:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Non concurrent refresh materialized view",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Refresh the materialized view in concurrent mode",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.PartitionCmd) -> None:
        """Fix violation."""
        node.concurrent = True



"""Non concurrent index drop."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class NonConcurrentIndexDrop(linter.BaseChecker):
    """## **What it does**
    Checks non-concurrent index drop.

    ## **Why not?**
    Dropping an index in non-concurrent mode acquires an **ACCESS EXCLUSIVE** lock on the
    table, blocking other accesses until the index drop is completed.
    This will cause downtime if the table is concurrently being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    Drop the index in concurrent mode: **DROP INDEX CONCURRENTLY ..**.
    """

    is_auto_fixable: bool = True

    def visit_DropStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.DropStmt,
    ) -> None:
        """Visit DropStmt."""
        if node.removeType == enums.ObjectType.OBJECT_INDEX and not node.concurrent:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Non concurrent index drop",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Drop the index in concurrent mode",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.DropStmt) -> None:
        """Fix violation."""
        node.concurrent = True



"""Checker for yoda conditions."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class YodaCondition(linter.BaseChecker):
    """## **What it does**
    Checks for yoda conditions.

    ## **Why not?**
    Yoda conditions can be harder to read and understand, especially for those who are not
    familiar with this syntax.

    Placing the constant on the left side makes it harder for someone to quickly grasp the
    meaning of the condition.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Natural condition.
    """

    is_auto_fixable: bool = True

    def visit_A_Expr(
        self,
        ancestors: visitors.Ancestor,
        node: ast.A_Expr,
    ) -> None:
        """Visit A_Expr."""
        if (
            node.kind == enums.A_Expr_Kind.AEXPR_OP
            and isinstance(
                node.lexpr,
                ast.A_Const,
            )
            and isinstance(node.rexpr, ast.ColumnRef)
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Yoda conditions are discouraged",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use natural condition",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.A_Expr) -> None:
        """Fix violation."""
        lexpr = node.lexpr
        rexpr = node.rexpr

        node.lexpr = rexpr
        node.rexpr = lexpr



"""Checker for rule."""

from pglast import ast, visitors

from pgrubic.core import linter


class CreateRule(linter.BaseChecker):
    """## **What it does**
    Checks for creation of rules.

    ## **Why not?**
    Rules are incredibly powerful, but they don't do what they look like they do.
    They look like they're some conditional logic, but they actually rewrite a query
    to modify it or add additional queries to it.

    ## **When should you?**
    Never. While the rewriter is an implementation detail of VIEWs,
    there is no reason to pry up this cover plate directly.

    ## **Use instead:**
    Don't use rules. If you think you want to, use a trigger instead.
    """

    def visit_RuleStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.RuleStmt,
    ) -> None:
        """Visit RuleStmt."""
        self.violations.add(
            linter.Violation(
                rule_code=self.code,
                rule_name=self.name,
                rule_category=self.category,
                line_number=self.line_number,
                column_offset=self.column_offset,
                line=self.line,
                statement_location=self.statement_location,
                description="Create rule detected",
                is_auto_fixable=self.is_auto_fixable,
                is_fix_enabled=self.is_fix_enabled,
                help="Use trigger",
            ),
        )



"""Checker for usage of NOT IN."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class NotIn(linter.BaseChecker):
    """## **What it does**
    Checks for usage of NOT IN.

    ## **Why not?**
    Don't use NOT IN, or any combination of NOT and IN such as NOT (x IN (select…))

    Two reasons:

    1. NOT IN behaves in unexpected ways if there is a null present:
    ```sql
    select * from foo where col not in (1,null); -- always returns 0 rows
    ```
    ```sql
    select * from foo where foo.col not in (select bar.x from bar); -- returns 0 rows if
    any value of bar.x is null
    ```

        This happens because col IN (1,null) returns TRUE if col=1, and NULL otherwise
        (i.e. it can never return FALSE). Since NOT (TRUE) is FALSE, but NOT (NULL) is
        still NULL, there is no way that NOT (col IN (1,null)) (which is the same thing
        as col NOT IN (1,null)) can return TRUE under any circumstances.

    2. Because of point 1 above, NOT IN (SELECT ...) does not optimize very well.
    In particular, the planner can't transform it into an anti-join, and so it becomes
    either a hashed Subplan or a plain Subplan. The hashed subplan is fast, but the
    planner only allows that plan for small result sets; the plain subplan is horrifically
    slow (in fact O(N²)). This means that the performance can look good in small-scale
    tests but then slow down by 5 or more orders of magnitude once a size threshold is
    crossed; you do not want this to happen.

    ## **When should you?**
    NOT IN (list,of,values,...) is mostly safe unless you might have a null in the list
    (via a parameter or otherwise). So it's sometimes natural and even advisable to use
    it when excluding specific constant values from a query result.

    ## **Use instead:**
    In most cases, the NULL behavior of NOT IN (SELECT …) is not intentionally desired,
    and the query can be rewritten using NOT EXISTS (SELECT …):
    ```sql
    select * from foo where not exists (select from bar where foo.col = bar.x);
    ```
    """

    def visit_A_Expr(
        self,
        ancestors: visitors.Ancestor,
        node: ast.A_Expr,
    ) -> None:
        """Visit A_Expr."""
        if node.kind == enums.A_Expr_Kind.AEXPR_IN and node.name[-1].sval == "<>":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="NOT IN detected",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use NOT EXISTS instead",
                ),
            )



"""Checker for varchar."""

from pglast import ast, visitors

from pgrubic.core import linter


class Varchar(linter.BaseChecker):
    """## **What it does**
    Checks for usage of varchar.

    ## **Why not?**
    varchar(n) is a variable width text field that will throw an error if you try and
    insert a string longer than n characters (not bytes) into it.

    varchar (without the (n)) or text are similar, but without the length limit.
    If you insert the same string into the three field types they will take up exactly
    the same amount of space, and you won't be able to measure any difference
    in performance.

    If what you really need is a text field with an length limit then varchar(n) is
    great, but if you pick an arbitrary length and choose varchar(20) for a surname
    field you are risking production errors in the future when Hubert Blaine
    Wolfe­schlegel­stein­hausen­berger­dorff signs up for your service.

    Some databases do not have a type that can hold arbitrary long text, or if they do
    it's not as convenient or efficient or well-supported as varchar(n).
    Users from those databases will often use something like varchar(255) when what
    they really want is text.

    If you need to constrain the value in a field you probably need something more
    specific than a maximum length - maybe a minimum length too, or a limited set of
    characters - and a check constraint can do all of those things as well as a maximum
    string length.

    ## **When should you?**
    When you want to, really. If what you want is a text field that will throw an error
    if you insert too long a string into it, and you don't want to use an explicit check
    constraint then varchar(n) is a perfectly good type. Just don't use it automatically
    without thinking about it.

    Also, the varchar type is in the SQL standard, unlike the text type, so it might be
    the best choice for writing super-portable applications.

    ## **Use instead:**
    1. text
    2. text with a constraint that enforces a maximum string length
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval == "varchar":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer text to varchar",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use text",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "text",
                },
            ),
        )



"""Formatter."""

import typing

from pglast import parser, stream

from pgrubic.core import noqa, config, errors


class FormatResult(typing.NamedTuple):
    """Format Result."""

    source_file: str
    original_source_code: str
    formatted_source_code: str


class Formatter:
    """Format source code."""

    def __init__(
        self,
        *,
        config: config.Config,
        formatters: typing.Callable[[], set[typing.Callable[[], None]]],
    ) -> None:
        """Initialize variables."""
        self.formatters = formatters()
        self.config = config

    @staticmethod
    def run(*, source_file: str, source_code: str, config: config.Config) -> str:
        """Format source code.

        Parameters:
        ----------
        source_file: str
            Path to the source file.
        source_code: str
            Source code to format.

        Returns:
        -------
        str
            Formatted source code.

        Raises:
        ------
        ParseError
            If there is an error parsing source code.
        """
        try:
            parser.parse_sql(source_code)

        except parser.ParseError as error:
            msg = "Error parsing source code: "
            raise errors.ParseError(f"{source_file}: " + msg + str(error)) from error

        formatted_statements: list[str] = []

        format_ignores = noqa.extract_format_ignores(
            source_file=source_file,
            source_code=source_code,
        )

        for statement in noqa.extract_statement_locations(
            source_file=source_file,
            source_code=source_code,
        ):
            if statement.start_location in format_ignores:
                formatted_statements.append(statement.text)
                continue

            comments = noqa.extract_comments(
                source_file=source_file,
                source_code=statement.text,
            )

            formatted_statement = stream.IndentedStream(
                comments=comments,
                semicolon_after_last_statement=False,
                remove_pg_catalog_from_functions=config.format.remove_pg_catalog_from_functions,
                comma_at_eoln=not (config.format.comma_at_beginning),
                special_functions=True,
            )(statement.text)

            if config.format.new_line_before_semicolon:
                formatted_statement += noqa.NEW_LINE + noqa.SEMI_COLON
            else:
                formatted_statement += noqa.SEMI_COLON

            formatted_statements.append(formatted_statement)

        return (
            noqa.NEW_LINE + (noqa.NEW_LINE * config.format.lines_between_statements)
        ).join(
            formatted_statements,
        ) + noqa.NEW_LINE

    def format(self, *, source_file: str, source_code: str) -> FormatResult:
        """Format source code.

        Parameters:
        ----------
        source_file: str
            Path to the source file.
        source_code: str
            Source code to format.

        Returns:
        -------
        FormatResult
            Formatted source code.
        """
        return FormatResult(
            source_file=source_file,
            original_source_code=source_code,
            formatted_source_code=self.run(
                source_file=source_file,
                source_code=source_code,
                config=self.config,
            ),
        )



"""Checker for unlogged table."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class UnloggedTable(linter.BaseChecker):
    """## **What it does**
    Checks for use of unlogged tables.

    ## **Why not?**
    Unlogged tables are not crash-safe: an unlogged table is automatically truncated after
    a crash or unclean shutdown. The contents of an unlogged table are also not replicated
    to standby servers.
    Any indexes created on an unlogged table are automatically unlogged as well. Any
    sequences created together with the unlogged table (for identity or serial columns)
    are also created as unlogged.

    ## **When should you?**
    The table is transient and its content can be regenerated after a crash or unclean
    shutdown.

    ## **Use instead:**
    Use a regular table.
    """

    is_auto_fixable: bool = True

    description: str = "Prefer regular table to unlogged table"
    help: str = "Use regular table"

    def visit_CreateStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateStmt,
    ) -> None:
        """Visit CreateStmt."""
        if node.relation.relpersistence == enums.RELPERSISTENCE_UNLOGGED:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=self.description,
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help=self.help,
                ),
            )

            self._fix_create_unlogged_table(node)

    def _fix_create_unlogged_table(self, node: ast.CreateStmt) -> None:
        """Fix violation."""
        node.relation.relpersistence = enums.RELPERSISTENCE_PERMANENT

    def visit_AlterTableCmd(
        self,
        ancestors: visitors.Ancestor,
        node: ast.AlterTableCmd,
    ) -> None:
        """Visit AlterTableCmd."""
        if node.subtype == enums.AlterTableType.AT_SetUnLogged:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=self.description,
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help=self.help,
                ),
            )

            self._fix_alter_unlogged_table(node)

    def _fix_alter_unlogged_table(self, node: ast.AlterTableCmd) -> None:
        """Fix violation."""
        node.subtype = enums.AlterTableType.AT_SetLogged



"""Checker for objects that are schema-qualifiable but are not schema qualified."""

from pglast import ast, visitors

from pgrubic.core import linter

SCHEMA_QUALIFIED_LENGTH = 2


class SchemaUnqualifiedObject(linter.BaseChecker):
    """## **What it does**
    Checks for objects that are schema-qualifiable but are not schema qualified.

    We currently do not check **subqueries**.

    ## **Why not?**
    Explicitly specifying schema improves code readability and improves clarity.

    ## **When should you?**
    If you really do not want to specify schema.

    ## **Use instead:**
    Specify schema.
    """

    help: str = "Schema qualify the object"

    def _check_enum_for_schema(
        self,
        node: ast.CreateEnumStmt | ast.AlterEnumStmt,
    ) -> None:
        """Check enum for schema."""
        if len(node.typeName) < SCHEMA_QUALIFIED_LENGTH:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Database object `{node.typeName[0].sval}`"
                    " should be schema qualified",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help=self.help,
                ),
            )

    def _check_function_for_schema(
        self,
        function_name: tuple[ast.String, ...],
    ) -> None:
        """Check function for schema."""
        if len(function_name) < SCHEMA_QUALIFIED_LENGTH:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Database object `{function_name[0].sval}`"
                    " should be schema qualified",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help=self.help,
                ),
            )

    def visit_RangeVar(
        self,
        ancestors: visitors.Ancestor,
        node: ast.RangeVar,
    ) -> None:
        """Visit RangeVar."""
        # Since CTEs are not schema qualifiable, they need to be excluded
        ctenames = []

        # we exclude referenced CTE names in CTEs
        if ancestors.find_nearest(ast.WithClause):
            ctenames = [
                cte.ctename for cte in ancestors.find_nearest(ast.WithClause).node.ctes
            ]

        # we exclude referenced CTE names in outer queries
        if (
            ancestors.find_nearest(
                ast.SelectStmt
                | ast.UpdateStmt
                | ast.DeleteStmt
                | ast.InsertStmt
                | ast.MergeStmt,
            )
            and ancestors.find_nearest(
                ast.SelectStmt
                | ast.UpdateStmt
                | ast.DeleteStmt
                | ast.InsertStmt
                | ast.MergeStmt,
            ).node.withClause
        ):
            ctenames = [
                cte.ctename
                for cte in ancestors.find_nearest(
                    ast.SelectStmt
                    | ast.UpdateStmt
                    | ast.DeleteStmt
                    | ast.InsertStmt
                    | ast.MergeStmt,
                ).node.withClause.ctes
            ]

        if (
            # there is no simple way to figure out if a subquery is referencing a CTE name
            # hence we are excluding all subqueries
            not (ancestors.find_nearest(ast.RangeSubselect | ast.SubLink))
            and node.relname not in ctenames
            and not node.schemaname
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Database object `{node.relname}`"
                    " should be schema qualified",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help=self.help,
                ),
            )

    def visit_DropStmt(
        self,
        ancestors: ast.Node,
        node: ast.DropStmt,
    ) -> None:
        """Visit DropStmt."""
        for obj in node.objects:
            object_names = getattr(obj, "names", getattr(obj, "objname", obj))

            if (
                isinstance(object_names, tuple | list)
                and len(object_names) < SCHEMA_QUALIFIED_LENGTH
            ):
                self.violations.add(
                    linter.Violation(
                        rule_code=self.code,
                        rule_name=self.name,
                        rule_category=self.category,
                        line_number=self.line_number,
                        column_offset=self.column_offset,
                        line=self.line,
                        statement_location=self.statement_location,
                        description=f"Database object `{object_names[-1].sval}`"
                        " should be schema qualified",
                        is_auto_fixable=self.is_auto_fixable,
                        is_fix_enabled=self.is_fix_enabled,
                        help=self.help,
                    ),
                )

    def visit_CreateEnumStmt(
        self,
        ancestors: ast.Node,
        node: ast.CreateEnumStmt,
    ) -> None:
        """Visit CreateEnumStmt."""
        self._check_enum_for_schema(node)

    def visit_AlterEnumStmt(
        self,
        ancestors: ast.Node,
        node: ast.AlterEnumStmt,
    ) -> None:
        """Visit AlterEnumStmt."""
        self._check_enum_for_schema(node)

    def visit_CreateFunctionStmt(
        self,
        ancestors: ast.Node,
        node: ast.CreateFunctionStmt,
    ) -> None:
        """Visit CreateFunctionStmt."""
        self._check_function_for_schema(node.funcname)

    def visit_AlterFunctionStmt(
        self,
        ancestors: ast.Node,
        node: ast.AlterFunctionStmt,
    ) -> None:
        """Visit AlterFunctionStmt."""
        self._check_function_for_schema(node.func.objname)

    def visit_ObjectWithArgs(
        self,
        ancestors: ast.Node,
        node: ast.ObjectWithArgs,
    ) -> None:
        """Visit ObjectWithArgs."""
        if len(node.objname) < SCHEMA_QUALIFIED_LENGTH:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Database object `{node.objname[0].sval}`"
                    " should be schema qualified",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help=self.help,
                ),
            )



"""Formatter for constraint."""

from pglast import ast, enums, stream, printers

from pgrubic import Operators


@printers.node_printer(ast.Constraint, override=True)
def constraint(node: ast.Constraint, output: stream.RawStream) -> None:
    """Printer for Constraint."""
    if node.conname:
        output.swrite("CONSTRAINT")
        output.space()
        output.print_name(node.conname)

    # Print the constraint definition
    printers.ddl.constr_type_printer(node.contype, node, output)

    if node.indexname:
        output.space()
        output.write("USING INDEX")
        output.space()
        output.print_name(node.indexname)

    if node.keys and node.contype in (
        enums.ConstrType.CONSTR_UNIQUE,
        enums.ConstrType.CONSTR_PRIMARY,
    ):
        output.space()
        with output.expression(need_parens=True):
            output.print_name(node.keys, ",")

    if node.including:
        output.space()
        output.write("INCLUDE")
        output.space()
        with output.expression(need_parens=True):
            output.print_list(node.including, ",", are_names=True)

    if node.deferrable:
        output.space()
        output.write("DEFERRABLE")
        if node.initdeferred:
            output.swrite("INITIALLY DEFERRED")

    if node.options:
        output.space()
        output.write("WITH")
        output.space()
        with output.expression(need_parens=True):
            output.print_list(node.options)

    if node.indexspace:
        output.space()
        output.writes("USING INDEX TABLESPACE")
        output.print_name(node.indexspace)

    if node.skip_validation:
        output.swrite("NOT VALID")


@printers.node_printer(ast.Constraint, ast.DefElem, override=True)
def constraint_def_elem(node: ast.DefElem, output: stream.RawStream) -> None:
    """Printer for Constraint defelem."""
    output.print_name(node.defname)
    if node.arg:
        output.write(Operators.EQ)
        output.space()
        output.print_node(node.arg)



"""Formatter for JOIN statements."""

from pglast import ast, enums, stream, printers


@printers.node_printer(ast.JoinExpr, override=True)
def join_expr(node: ast.JoinExpr, output: stream.RawStream) -> None:
    """Printer for JoinExpr."""
    indent = (
        -6
        if node.jointype in (enums.JoinType.JOIN_RIGHT, enums.JoinType.JOIN_INNER)
        else -5
    )

    if node.isNatural:
        indent -= 1

    with output.push_indent(indent):
        with output.expression(bool(node.alias)):
            output.print_node(node.larg)
            output.newline()

            if node.isNatural:
                output.write("NATURAL")
                output.space()

            if node.jointype == enums.JoinType.JOIN_INNER:
                if not node.usingClause and not node.quals and not node.isNatural:
                    output.write("CROSS")
                    output.space()
                else:
                    output.write("INNER")
                    output.space()
            elif node.jointype == enums.JoinType.JOIN_LEFT:
                output.write("LEFT")
                output.space()
            elif node.jointype == enums.JoinType.JOIN_FULL:
                output.write("FULL")
                output.space()
            elif node.jointype == enums.JoinType.JOIN_RIGHT:
                output.write("RIGHT")
                output.space()

            output.swrite("JOIN")
            output.space()

            if isinstance(node.rarg, ast.JoinExpr):
                output.indent(3, relative=False)
                # TODO: This needs to be fixed to handle nested JOIN expressions properly  # noqa: TD002, FIX002, TD003, E501
                with output.expression(not bool(node.rarg.alias)):
                    output.indent(4)
                    output.print_node(node.rarg)
                output.newline()
            else:
                output.print_node(node.rarg)

            if node.usingClause:
                output.swrite("USING")
                output.space()
                with output.expression(need_parens=True):
                    output.print_name(node.usingClause, ",")
                if node.join_using_alias:
                    output.space()
                    output.write("AS")
                    output.space()
                    output.print_node(node.join_using_alias)
            elif node.quals:
                output.newline()
                if node.jointype in (
                    enums.JoinType.JOIN_RIGHT,
                    enums.JoinType.JOIN_INNER,
                ):
                    output.space(3)
                else:
                    output.space(2)
                output.swrite("ON")
                output.space()
                output.print_node(node.quals)

        if node.alias:
            output.space()
            output.writes("AS")
            output.space()
            output.print_name(node.alias)

        if isinstance(node.rarg, ast.JoinExpr):
            output.dedent()



"""pgrubic errors."""


class BaseError(Exception):
    """Base class for all exceptions."""


class MissingConfigError(BaseError):
    """Raised when a config is missing."""


class ParseError(BaseError):
    """Raised when a parse error occurs."""


class MissingStatementTerminatorError(BaseError):
    """Raised when a statement terminator is missing."""



"""Handling noqa comments."""

import sys
import typing
import pathlib
import dataclasses

from pglast import parser
from colorama import Fore, Style

from pgrubic import PACKAGE_NAME
from pgrubic.core import errors

A_STAR: str = "*"
ASCII_SEMI_COLON: str = "ASCII_59"
SEMI_COLON: str = ";"
NEW_LINE: str = "\n"


class Statement(typing.NamedTuple):
    """Representation of an SQL statement."""

    start_location: int
    end_location: int
    text: str


def extract_statement_locations(
    *,
    source_file: str,
    source_code: str,
) -> list[Statement]:
    """Build statements start and end locations.

    Parameters:
    ----------
    source_file: str
        Path to the source file.
    source_code: str
        Source code to lint.

    Returns:
    -------
    list[Statement]
        List of statements.

    Raises:
    ------
    MissingStatementTerminatorError
        If semicolon is missing at the end of a statement.
    """
    locations: list[Statement] = []

    statement_start_location = 0

    tokens = parser.scan(source_code)

    for idx, token in enumerate(tokens):
        if idx == len(tokens) - 1 and token.name != ASCII_SEMI_COLON:
            msg = f"Missing semicolon (;) at end of SQL statement at location {token.end}"

            raise errors.MissingStatementTerminatorError(f"{source_file}: " + msg)

        if token.name == ASCII_SEMI_COLON:
            locations.append(
                Statement(
                    start_location=statement_start_location,
                    end_location=token.end,
                    text=source_code[statement_start_location : token.end] + SEMI_COLON,
                ),
            )
            statement_start_location = token.end + 1

    return locations


def _get_rules_from_inline_comment(
    comment: str,
    location: int,
    section: str = "noqa",
) -> list[str]:
    """Get rules from inline comment.

    Parameters:
    ----------
    comment: str
        Inline comment.
    location: int
        Location of inline comment.
    section: str
        Section of inline comment.

    Returns:
    -------
    list[str]
        List of rules.
    """
    comment_remainder = comment.removeprefix(section)

    if not comment_remainder:
        return [A_STAR]

    rules: list[str] = [
        rule.strip()
        for rule in comment_remainder.removeprefix(":").split(",")
        if rule and comment_remainder.startswith(":")
    ]

    if not rules:
        sys.stderr.write(
            f"{Fore.YELLOW}Warning: Malformed `noqa` directive at location {location}. Expected `noqa: <rules>`{Style.RESET_ALL}\n",  # noqa: E501
        )

    return rules


def _get_statement_locations(
    locations: list[Statement],
    stop: int,
) -> tuple[int, int]:
    """Get statement start and end locations.

    Parameters:
    ----------
    locations: list[Statement]
        List of statements.
    stop: int
        Stop location.

    Returns:
    -------
    tuple[int, int]
        Statement start and end locations.
    """
    for statement_start_location, statement_end_location, _ in locations:
        if statement_start_location <= stop < statement_end_location:
            break

    return statement_start_location, statement_end_location


@dataclasses.dataclass(kw_only=True)
class NoQaDirective:
    """Representation of a noqa directive."""

    source_file: str | None = None
    location: int
    line_number: int
    column_offset: int
    rule: str
    used: bool = False


def _extract_statement_ignores(
    source_file: str,
    source_code: str,
) -> list[NoQaDirective]:
    """Extract ignores from SQL statements.

    Parameters:
    ----------
    source_file: str
        Path to the source file.
    source_code: str
        Source code to lint.

    Returns:
    -------
    list[NoQaDirective]
        List of ignores.
    """
    locations = extract_statement_locations(
        source_file=source_file,
        source_code=source_code,
    )

    inline_ignores: list[NoQaDirective] = []

    for token in parser.scan(source_code):
        if token.name == "SQL_COMMENT":
            statement_start_location, statement_end_location = _get_statement_locations(
                locations,
                token.start,
            )

            line_number = source_code[:statement_end_location].count("\n") + 1

            # Here, we extract last comment because we can have a comment followed
            # by another comment e.g -- new table -- noqa: US005
            comment = source_code[token.start : (token.end + 1)].split("--")[-1].strip()

            if comment.startswith("noqa"):
                rules = _get_rules_from_inline_comment(comment, token.start)

                inline_ignores.extend(
                    NoQaDirective(
                        location=statement_start_location,
                        line_number=line_number,
                        column_offset=(statement_end_location - token.start),
                        rule=rule,
                    )
                    for rule in rules
                )

    return inline_ignores


def _extract_file_ignore(source_file: str, source_code: str) -> list[NoQaDirective]:
    """Extract ignore from the start of a source file.

    Parameters:
    ----------
    source_file: str
        Path to the source file.
    source_code: str
        Source code to lint.

    Returns:
    -------
    list[NoQaDirective]
        List of ignores.
    """
    file_ignores: list[NoQaDirective] = []

    for token in parser.scan(source_code):
        if token.start == 0 and token.name == "SQL_COMMENT":
            comment = source_code[token.start : (token.end + 1)].split("--")[-1].strip()

            if comment.strip().startswith(f"{PACKAGE_NAME}: noqa"):
                rules = _get_rules_from_inline_comment(
                    comment,
                    token.start,
                    section=f"{PACKAGE_NAME}: noqa",
                )

                file_ignores.extend(
                    NoQaDirective(
                        source_file=source_file,
                        location=token.start,
                        line_number=1,
                        column_offset=0,
                        rule=rule,
                    )
                    for rule in rules
                )
        else:
            break

    return file_ignores


def extract_ignores(*, source_file: str, source_code: str) -> list[NoQaDirective]:
    """Extract ignores from source code.

    Parameters:
    ----------
    source_file: str
        Path to the source file.
    source_code: str
        Source code to lint.

    Returns:
    -------
    list[NoQaDirective]
        List of ignores.
    """
    return _extract_statement_ignores(
        source_file=source_file,
        source_code=source_code,
    ) + _extract_file_ignore(
        source_file=source_file,
        source_code=source_code,
    )


def extract_format_ignores(source_file: str, source_code: str) -> list[int]:
    """Extract format ignores from SQL statements.

    Parameters:
    ----------
    source_file: str
        Path to the source file.
    source_code: str
        Source code to lint.

    Returns:
    -------
    list[int]
        List of ignores.
    """
    locations = extract_statement_locations(
        source_file=source_file,
        source_code=source_code,
    )

    inline_ignores: list[int] = []

    for token in parser.scan(source_code):
        if token.name == "SQL_COMMENT":
            statement_start_location, _ = _get_statement_locations(
                locations,
                token.start,
            )

            comment = source_code[token.start : (token.end + 1)].split("--")[-1].strip()

            if (
                comment.strip().startswith("fmt")
                and comment.removeprefix("fmt").removeprefix(":").strip() == "skip"
            ):
                inline_ignores.append(
                    statement_start_location,
                )

    return inline_ignores


class Comment(typing.NamedTuple):
    """Representation of an SQL comment."""

    location: int
    text: str
    at_start_of_line: bool
    continue_previous: bool


def extract_comments(*, source_file: str, source_code: str) -> list[Comment]:
    """Extract comments from SQL statements.

    Parameters:
    ----------
    source_file: str
        Path to the source file.
    source_code: str
        Source code to lint.

    Returns:
    -------
    list[Comment]
        List of comments.
    """
    locations = extract_statement_locations(
        source_file=source_file,
        source_code=source_code,
    )

    comments: list[Comment] = []
    continue_previous = False

    for token in parser.scan(source_code):
        if token.name in ("C_COMMENT", "SQL_COMMENT"):
            statement_start_location, _ = _get_statement_locations(
                locations,
                token.start,
            )

            comment = source_code[token.start : (token.end + 1)]
            at_start_of_line = not source_code[
                : token.start - statement_start_location
            ].strip()
            comment = source_code[token.start : token.end + 1]
            comments.append(
                Comment(
                    statement_start_location,
                    comment,
                    at_start_of_line,
                    continue_previous,
                ),
            )
            continue_previous = True
        else:
            continue_previous = False
    return comments


def report_unused_ignores(
    *,
    source_file: str,
    inline_ignores: list[NoQaDirective],
) -> None:
    """Get unused ignores.

    Parameters:
    ----------
    source_file: str
        Path to the source file.
    inline_ignores: list[NoQaDirective]
        Inline noqa directives.

    Returns:
    -------
    None
    """
    for ignore in inline_ignores:
        if not ignore.used:
            sys.stdout.write(
                f"{source_file}:{ignore.line_number}:{ignore.column_offset}:"
                f" {Fore.YELLOW}Unused noqa directive{Style.RESET_ALL}"
                f" (unused: {Fore.RED}{Style.BRIGHT}{ignore.rule}{Style.RESET_ALL})\n",
            )


def add_file_level_general_ignore(sources: set[pathlib.Path]) -> int:
    """Add file-level general noqa directive to the begining of each source.

    Parameters:
    ----------
    sources: set[pathlib.Path]
        Set of source files.

    Returns:
    -------
    int
        Number of sources modified.
    """
    sources_modified = 0

    for source in sources:
        skip = False
        source_code = source.read_text()

        for token in parser.scan(source_code):
            if token.start == 0 and token.name == "SQL_COMMENT":
                comment = (
                    source_code[token.start : (token.end + 1)].split("--")[-1].strip()
                )

                if comment.strip() == f"{PACKAGE_NAME}: noqa":
                    skip = True
                    break

        if not skip:
            source.write_text(f"-- {PACKAGE_NAME}: noqa\n{source_code}")
            sources_modified += 1
            continue

    return sources_modified



"""Test loader."""

from pgrubic import core


def test_load_rules(linter: core.Linter) -> None:
    """Test loading rules."""
    expected_number_of_rules = 108

    rules = core.load_rules(config=linter.config)

    assert len(rules) == expected_number_of_rules



"""Checker for duplicate columns."""

from pglast import ast, visitors

from pgrubic.core import linter
from pgrubic.rules.general import get_columns_from_table_creation


class DuplicateColumn(linter.BaseChecker):
    """## **What it does**
    Checks for duplicate columns.

    ## **Why not?**
    Each column in a table must have a unique name within that table to ensure unambiguous
    reference and avoid confusion during SQL operations. It ensures data integrity and
    clarity in database design and manipulation.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Remove duplicate columns.
    """

    def visit_CreateStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateStmt,
    ) -> None:
        """Visit CreateStmt."""
        _, duplicate_columns = get_columns_from_table_creation(node)

        for column in duplicate_columns:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Column `{column}` specified more than once in table {node.relation.relname}",  # noqa: E501
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help=f"Remove duplicate column from table {node.relation.relname}",
                ),
            )



"""Checker for duplicate index column."""

from pglast import ast, visitors

from pgrubic.core import linter


class DuplicateIndexColumn(linter.BaseChecker):
    """## **What it does**
    Checks for duplicate index column.

    ## **Why not?**
    Having duplicate column in an index is redundant and unnecessary.
    Indexes are not free, they add an overhead to write operations,
    and having a column repeated does not add any benefit but rather unnecessary
    index maintenance and storage.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Remove duplicate column from the index.
    """

    is_auto_fixable: bool = True

    def visit_IndexStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.IndexStmt,
    ) -> None:
        """Visit IndexStmt."""
        columns: list[str] = [column.name for column in node.indexParams if column.name]
        duplicate_columns = {column for column in columns if columns.count(column) > 1}

        for column in duplicate_columns:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Column `{column}` specified more than once in index",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help=f"Remove duplicate column `{column}` from index",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.IndexStmt) -> None:
        """Fix violation."""
        index_params: list[ast.IndexElem] = []

        for param in node.indexParams:
            if param.name not in [column.name for column in index_params]:
                index_params.append(param)

        node.indexParams = index_params



"""Checker for drop table."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class DropTable(linter.BaseChecker):
    """## **What it does**
    Checks drop table.

    ## **Why not?**
    Not only that mistakenly dropping a table can cause data loss, applications that
    rely on the data will break.

    ## **When should you?**
    If you really want to drop the table.

    ## **Use instead:**
    No suggestions.
    """

    def visit_DropStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.DropStmt,
    ) -> None:
        """Visit DropStmt."""
        if node.removeType == enums.ObjectType.OBJECT_TABLE:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Drop table found",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                ),
            )



"""Checker for existence of not null constraint on required columns."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class NullableRequiredColumn(linter.BaseChecker):
    """## **What it does**
    Checks for existence of not null constraint on required columns.

    ## **Why not?**
    If a column has been specified as required then it should not be nullable.
    Having a required column as nullable is an anti-pattern and should be avoided.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Set the required column as **Not Null**.

    ## **Configuration**
    `required-columns`: List of required columns along with their data types.
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        for column in self.config.lint.required_columns:
            if node.colname == column.name:
                is_not_null = bool(
                    (
                        [
                            constraint
                            for constraint in node.constraints
                            if constraint.contype == enums.ConstrType.CONSTR_NOTNULL
                        ]
                        if node.constraints is not None
                        else []
                    ),
                )

                if not is_not_null:
                    self.violations.add(
                        linter.Violation(
                            rule_code=self.code,
                            rule_name=self.name,
                            rule_category=self.category,
                            line_number=self.line_number,
                            column_offset=self.column_offset,
                            line=self.line,
                            statement_location=self.statement_location,
                            description=f"Column `{node.colname}` is marked as required"
                            " in config",
                            is_auto_fixable=self.is_auto_fixable,
                            is_fix_enabled=self.is_fix_enabled,
                            help="Set the required column as Not Null",
                        ),
                    )

                    self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.constraints = (
            *(node.constraints or []),
            ast.Constraint(
                contype=enums.ConstrType.CONSTR_NOTNULL,
            ),
        )



"""Checker for unlogged table."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class CurrentTime(linter.BaseChecker):
    """## **What it does**
    Checks for use of **CURRENT_TIME** function.

    ## **Why not?**
    It returns a value of type **time with time zone**.

    The manual has more to say about this type:

    > The type time with time zone is defined by the SQL standard, but the definition
    > exhibits properties which lead to questionable usefulness. In most cases,
    > a combination of date, time, timestamp without time zone, and
    > timestamp with time zone should provide a complete range of date/time
    > functionality required by any application.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Do not use the CURRENT_TIME function. Use whichever of these is appropriate:

    - CURRENT_TIMESTAMP or now() if you want a timestamp with time zone,
    - LOCALTIMESTAMP if you want a timestamp without time zone,
    - CURRENT_DATE if you want a date,
    - LOCALTIME if you want a time
    """

    is_auto_fixable: bool = True

    def visit_SQLValueFunction(
        self,
        ancestors: visitors.Ancestor,
        node: ast.SQLValueFunction,
    ) -> None:
        """Visit SQLValueFunction."""
        if node.op == enums.SQLValueFunctionOp.SVFOP_CURRENT_TIME:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer functions that return timestamptz"
                    " instead of timetz",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use functions that return timestamptz",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.SQLValueFunction) -> None:
        """Fix violation."""
        node.op = enums.SQLValueFunctionOp.SVFOP_CURRENT_TIMESTAMP



"""Checker for timestamp with time zone with precision."""

from pglast import ast, visitors

from pgrubic.core import linter


class TimestampWithTimezoneWithPrecision(linter.BaseChecker):
    """## **What it does**
    Checks for usage of timestamp with time zone with precision.

    ## **Why not?**
    Because it rounds off the fractional part rather than truncating it as everyone
    would expect. This can cause unexpected issues; consider that when you store now()
    into such a column, you might be storing a value half a second in the future.

    ## **When should you?**
    Never.

    ## **Use instead:**
    timestamptz (also known as timestamp with time zone) without precision.
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval == "timestamptz" and node.typeName.typmods:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer entire timestamp with timezone",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use timestamptz without precision",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "timestamptz",
                },
            ),
        )



"""Formatter for enum."""

from pglast import ast, stream, printers

from pgrubic.formatters.ddl import IF_NOT_EXISTS


@printers.node_printer(ast.CreateEnumStmt, override=True)
def create_enum_stmt(node: ast.Node, output: stream.RawStream) -> None:
    """Printer for CreateEnumStmt."""
    output.write("CREATE TYPE")
    output.space()
    output.print_name(node.typeName)
    output.write("AS ENUM")
    output.space()
    with output.expression(need_parens=True):
        output.newline()
        output.space(4)
        output.print_list(node.vals, standalone_items=True)
        output.newline()


@printers.node_printer(ast.AlterEnumStmt, override=True)
def alter_enum_stmt(node: ast.AlterEnumStmt, output: stream.RawStream) -> None:
    """Printer for AlterEnumStmt."""
    output.write("ALTER TYPE")
    output.space()
    output.print_name(node.typeName)
    output.newline()
    output.space(4)
    if node.newVal:
        if node.oldVal:
            output.write("RENAME VALUE")
            output.space()
            output.write_quoted_string(node.oldVal)
            output.space()
            output.write("TO")
            output.space()
        else:
            output.write("ADD VALUE")
            if node.skipIfNewValExists:
                output.space()
                output.write(IF_NOT_EXISTS)
            output.space()
        output.write_quoted_string(node.newVal)
    if node.newValNeighbor:
        if node.newValIsAfter:
            output.space()
            output.write("AFTER")
            output.space()
        else:
            output.space()
            output.write("BEFORE")
            output.space()
        output.write_quoted_string(node.newValNeighbor)



"""Test cli."""

import pathlib
from unittest.mock import patch

import pytest
from click import testing

from tests import TEST_FILE
from pgrubic import WORKERS_ENVIRONMENT_VARIABLE
from pgrubic.__main__ import cli


def test_cli_lint_file(tmp_path: pathlib.Path) -> None:
    """Test cli lint file."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail)])

    assert result.exit_code == 1


def test_cli_lint_directory(tmp_path: pathlib.Path) -> None:
    """Test cli lint directory."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(directory)])

    assert result.exit_code == 1


def test_cli_lint_current_directory(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test cli lint current directory."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()
    monkeypatch.chdir(directory)

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint"])

    assert result.exit_code == 1


def test_cli_lint_complete_fix(tmp_path: pathlib.Path) -> None:
    """Test cli lint complete fix."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail), "--fix"])

    assert result.exit_code == 0


def test_cli_lint_with_add_file_level_general_noqa(tmp_path: pathlib.Path) -> None:
    """Test cli lint with add_file_level_general_noqa."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail), "--add-file-level-general-noqa"])

    assert result.output == "File-level general noqa directive added to 1 file(s)\n"

    assert result.exit_code == 0


def test_cli_lint_verbose(tmp_path: pathlib.Path) -> None:
    """Test cli lint verbose."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail), "--verbose"])

    assert result.exit_code == 1


def test_cli_lint_partial_fix(tmp_path: pathlib.Path) -> None:
    """Test cli lint partial fix."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL; SELECT * FROM example;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail), "--fix"])

    assert result.exit_code == 1


def test_cli_lint_ignore_noqa(tmp_path: pathlib.Path) -> None:
    """Test cli lint ignore noqa."""
    runner = testing.CliRunner()

    sql_fail: str = """
    -- noqa: GN024
    SELECT a = NULL;
    """

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["lint", str(file_fail), "--ignore-noqa"])

    assert result.exit_code == 1


def test_cli_lint_parse_error(tmp_path: pathlib.Path) -> None:
    """Test cli lint parse error."""
    runner = testing.CliRunner()

    sql: str = "CREATE TABLE tbl (activated);"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    result = runner.invoke(cli, ["lint", str(file_fail)])

    expected_exit_code: int = 2

    assert result.exit_code == expected_exit_code


def test_cli_format_file(tmp_path: pathlib.Path) -> None:
    """Test cli format file."""
    runner = testing.CliRunner()

    sql_pass: str = "SELECT a = NULL;\n"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_pass = directory / TEST_FILE
    file_pass.write_text(sql_pass)

    result = runner.invoke(cli, ["format", str(file_pass)])

    assert result.output == "1 file(s) reformatted, 0 file(s) left unchanged\n"

    assert result.exit_code == 0


def test_cli_format_file_verbose(tmp_path: pathlib.Path) -> None:
    """Test cli format file."""
    runner = testing.CliRunner()

    sql_pass: str = "SELECT a = NULL;\n"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_pass = directory / TEST_FILE
    file_pass.write_text(sql_pass)

    result = runner.invoke(cli, ["format", str(file_pass), "--verbose"])

    assert "Using default settings" in result.output

    assert result.exit_code == 0


def test_cli_format_directory(tmp_path: pathlib.Path) -> None:
    """Test cli format directory."""
    runner = testing.CliRunner()

    sql_pass: str = "SELECT a = NULL;\n"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_pass = directory / TEST_FILE
    file_pass.write_text(sql_pass)

    result = runner.invoke(cli, ["format", str(directory)])

    assert result.output == "1 file(s) reformatted, 0 file(s) left unchanged\n"

    assert result.exit_code == 0


def test_cli_format_current_directory(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test cli format current directory."""
    runner = testing.CliRunner()

    sql_pass: str = "SELECT a = NULL; SELECT * FROM example;"

    directory = tmp_path / "sub"
    directory.mkdir()
    monkeypatch.chdir(directory)

    file_pass = directory / TEST_FILE
    file_pass.write_text(sql_pass)

    result = runner.invoke(cli, ["format"])

    assert result.output == "1 file(s) reformatted, 0 file(s) left unchanged\n"

    assert result.exit_code == 0


def test_cli_format_check(tmp_path: pathlib.Path) -> None:
    """Test cli format check."""
    runner = testing.CliRunner()

    sql_fail: str = "SELECT a = NULL; SELECT * FROM example;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql_fail)

    result = runner.invoke(cli, ["format", str(file_fail), "--check"])

    assert result.output == ""

    assert result.exit_code == 1


def test_cli_format_diff(tmp_path: pathlib.Path) -> None:
    """Test cli format check."""
    runner = testing.CliRunner()

    sql: str = "SELECT a = NULL; SELECT * FROM example;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    result = runner.invoke(cli, ["format", str(file_fail), "--diff"])

    assert result.exit_code == 1


def test_cli_format_no_cache(tmp_path: pathlib.Path) -> None:
    """Test cli format with no cache."""
    runner = testing.CliRunner()

    sql: str = "SELECT a = NULL; SELECT * FROM example;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    result = runner.invoke(cli, ["format", str(file_fail)])

    assert result.output == "1 file(s) reformatted, 0 file(s) left unchanged\n"

    assert result.exit_code == 0

    # with cache read
    result = runner.invoke(cli, ["format", str(file_fail)])

    assert result.output == "0 file(s) reformatted, 1 file(s) left unchanged\n"

    assert result.exit_code == 0

    # without cache
    result = runner.invoke(cli, ["format", str(file_fail), "--no-cache"])

    assert result.output == "1 file(s) reformatted, 0 file(s) left unchanged\n"

    assert result.exit_code == 0


def test_cli_format_parse_error(tmp_path: pathlib.Path) -> None:
    """Test cli format parse error."""
    runner = testing.CliRunner()

    sql: str = "SELECT * FROM;"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    result = runner.invoke(cli, ["format", str(file_fail)])

    expected_exit_code: int = 2

    assert result.exit_code == expected_exit_code


def test_cli_format_missing_terminator_error(tmp_path: pathlib.Path) -> None:
    """Test cli format missing terminator error."""
    runner = testing.CliRunner()

    sql: str = "SELECT * FROM tbl"

    directory = tmp_path / "sub"
    directory.mkdir()

    file_fail = directory / TEST_FILE
    file_fail.write_text(sql)

    result = runner.invoke(cli, ["format", str(file_fail)])

    expected_exit_code: int = 2

    assert result.exit_code == expected_exit_code


def test_max_workers_from_environment_variable(tmp_path: pathlib.Path) -> None:
    """Test max workers from environment variable."""
    with patch.dict(
        "os.environ",
        {WORKERS_ENVIRONMENT_VARIABLE: "1"},
    ):
        runner = testing.CliRunner()

        sql: str = "SELECT * FROM tbl;"

        directory = tmp_path / "sub"
        directory.mkdir()

        file_fail = directory / TEST_FILE
        file_fail.write_text(sql)

        result = runner.invoke(cli, ["format", str(file_fail)])

        assert result.exit_code == 0



"""Checker for json."""

from pglast import ast, visitors

from pgrubic.core import linter


class Json(linter.BaseChecker):
    """## **What it does**
    Checks for usage of json.

    ## **Why not?**
    From the manual:
    > The json and jsonb data types accept almost identical sets of values as input.
    > The major practical difference is one of efficiency. The json data type stores an
    > exact copy of the input text, which processing functions must reparse on each
    > execution; while jsonb data is stored in a decomposed binary format that makes it
    > slightly slower to input due to added conversion overhead, but significantly
    > faster to process, since no reparsing is needed. jsonb also supports indexing,
    > which can be a significant advantage.

    ## **When should you?**
    In general, most applications should prefer to store JSON data as jsonb, unless
    there are quite specialized needs, such as legacy assumptions about ordering of
    object keys.

    ## **Use instead:**
    jsonb.
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval == "json":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer jsonb over json",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use jsonb",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "jsonb",
                },
            ),
        )



"""Checker for rename table."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class RenameTable(linter.BaseChecker):
    """## **What it does**
    Checks renaming of table.

    ## **Why not?**
    Renaming a table can easily break applications that rely on the table.

    If any part of the application code, database procedures, views, or reports use
    the table, renaming it will cause errors and potentially disrupt business operations.

    ## **When should you?**
    If the table is no longer being accessed by clients, probably after migrating
    clients to a new table.

    ## **Use instead:**
    1. Create a new table with the new name.
    2. Start writing data to the new table.
    3. Copy all data from the old table to the new table.
    4. Migrate clients to the new table.
    """

    def visit_RenameStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.RenameStmt,
    ) -> None:
        """Visit RenameStmt."""
        if (
            node.renameType == enums.ObjectType.OBJECT_TABLE
            and node.newname != node.relation.relname
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Rename table detected",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Create a new table with the new name",
                ),
            )



"""Checker for not null constraint on existing column."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class NotNullConstraintOnExistingColumn(linter.BaseChecker):
    """## **What it does**
    Checks **NOT NULL** constraint on an existing column.

    ## **Why not?**
    Adding a **NOT NULL** constraint to an existing column of an already populated table
    will have to scan and validate that there are no nulls in the column, causing the
    table to be locked in which no other operations can be performed on the table for the
    duration of the validation. This will cause downtime if the table is concurrently
    being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    1. Create a check constraint: **CHECK (column IS NOT NULL) NOT VALID**.
    2. Validate the constraint.
    3. Set the column as NOT NULL.
    4. Drop the constraint.
    """

    def visit_AlterTableCmd(
        self,
        ancestors: visitors.Ancestor,
        node: ast.AlterTableCmd,
    ) -> None:
        """Visit AlterTableCmd."""
        if node.subtype == enums.AlterTableType.AT_SetNotNull:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Not null constraint on existing column `{node.name}`",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Add a supporting check constraint",
                ),
            )



"""Checker for duplicate column in unique key constraint."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class DuplicateUniqueKeyColumn(linter.BaseChecker):
    """## **What it does**
    Checks for duplicate column in unique key constraint.

    ## **Why not?**
    While PostgreSQL does not allow duplicate columns in a unique key constraint,
    such constructs are still parseable, but will error out at runtime.

    Having a duplicate column in a unique key constraint is an obvious error.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Remove duplicate column from the unique key constraint.
    """

    is_auto_fixable: bool = True

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if node.contype == enums.ConstrType.CONSTR_UNIQUE:
            columns: list[str] = [key.sval for key in node.keys or []]
            duplicate_columns = {
                column for column in columns if columns.count(column) > 1
            }

            for column in duplicate_columns:
                self.violations.add(
                    linter.Violation(
                        rule_code=self.code,
                        rule_name=self.name,
                        rule_category=self.category,
                        line_number=self.line_number,
                        column_offset=self.column_offset,
                        line=self.line,
                        statement_location=self.statement_location,
                        description=f"Column `{column}` specified more than once in unique key constraint",  # noqa: E501
                        is_auto_fixable=self.is_auto_fixable,
                        is_fix_enabled=self.is_fix_enabled,
                        help="Remove duplicate column from unique key constraint",
                    ),
                )

            self._fix(node)

    def _fix(self, node: ast.Constraint) -> None:
        """Fix violation."""
        keys: list[ast.IndexElem] = []

        for key in node.keys:
            if key.sval not in [column.sval for column in keys]:
                keys.append(key)

        node.keys = keys



"""Checker for table column conflict."""

from pglast import ast, enums, visitors

from pgrubic.core import linter
from pgrubic.rules.general import get_columns_from_table_creation


class TableColumnConflict(linter.BaseChecker):
    """## **What it does**
    Checks for table column conflict.

    ## **Why not?**
    While each column in a table must have a unique name within that table to ensure
    unambiguous reference and avoid confusion during SQL operations, the name of a table
    should also be distinct from the name of its columns to avoid confusion and ensures
    clarity in database design and manipulation.

    ## **When should you?**
    Almost never.

    ## **Use instead:**
    Resolve the name conflict.
    """

    def _register_violation(
        self,
        table_name: str,
        line_number: int,
        column_offset: int,
        line: str,
        statement_location: int,
    ) -> None:
        """Register the violation."""
        self.violations.add(
            linter.Violation(
                rule_code=self.code,
                rule_name=self.name,
                rule_category=self.category,
                line_number=line_number,
                column_offset=column_offset,
                line=line,
                statement_location=statement_location,
                description=f"Table name `{table_name}` conflicts with the"
                " name of its column(s)",
                is_auto_fixable=self.is_auto_fixable,
                is_fix_enabled=self.is_fix_enabled,
                help="Resolve the name conflict",
            ),
        )

    def visit_CreateStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateStmt,
    ) -> None:
        """Visit CreateStmt."""
        given_columns, _ = get_columns_from_table_creation(node)

        if any(column.name == node.relation.relname for column in given_columns):
            self._register_violation(
                table_name=node.relation.relname,
                line_number=self.line_number,
                column_offset=self.column_offset,
                line=self.line,
                statement_location=self.statement_location,
            )

    def visit_AlterTableStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.AlterTableStmt,
    ) -> None:
        """Visit AlterTableStmt."""
        given_columns: list[str] = [
            cmd.def_.colname
            for cmd in node.cmds
            if cmd.subtype == enums.AlterTableType.AT_AddColumn
        ]

        if node.relation.relname in given_columns:
            self._register_violation(
                table_name=node.relation.relname,
                line_number=self.line_number,
                column_offset=self.column_offset,
                line=self.line,
                statement_location=self.statement_location,
            )



"""Checker for index elements."""

from pglast import ast, visitors

from pgrubic.core import linter


class IndexElementsMoreThanThree(linter.BaseChecker):
    """## **What it does**
    Checks for indexes with more than three elements.

    ## **Why not?**
    From the documentation:
    > Multicolumn indexes should be used sparingly. In most situations, an index on a
    > single column is sufficient and saves space and time.
    > Indexes with more than three columns are unlikely to be helpful unless the usage of
    > the table is extremely stylized.

    ## **When should you?**
    If you really need to.

    See [indexes-bitmap-scans](https://www.postgresql.org/docs/current/indexes-bitmap-scans.html)
    for some discussion of the merits of different index configurations.

    ## **Use instead:**
    Keep your index elements at most three.
    """

    def visit_IndexStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.IndexStmt,
    ) -> None:
        """Visit IndexStmt."""
        max_index_elements = 3

        if len(node.indexParams) > max_index_elements:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Index elements more than {max_index_elements}",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Keep index elements at most three",
                ),
            )



"""Checker for time with time zone."""

from pglast import ast, visitors

from pgrubic.core import linter


class TimeWithTimeZone(linter.BaseChecker):
    """## **What it does**
    Checks for usage of time with time zone.

    ## **Why not?**
    Even the manual tells you it is only implemented for SQL compliance:

    > The type time with time zone is defined by the SQL standard, but the definition
    > exhibits properties which lead to questionable usefulness. In most cases,
    > a combination of date, time, timestamp without time zone, and
    > timestamp with time zone should provide a complete range of date/time
    > functionality required by any application.

    ## **When should you?**
    Never.

    ## **Use instead:**
    timestamptz (also known as timestamp with time zone).
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval == "timetz":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer timestamp with timezone over time with timezone",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use timestamptz",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "timestamptz",
                },
            ),
        )



"""Checker for removal of constraints."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class RemoveConstraint(linter.BaseChecker):
    """## **What it does**
    Checks for removal of constraints.

    ## **Why not?**
    Constraints are used to ensure data quality, data integrity, consistency,
    and adherence to certain business rules.

    If a constraint is removed without due consideration, the data in the database can
    become corrupted.

    ## **When should you?**
    You can remove constraints if they are no longer needed or need adjustment. Don't just
    remove them without due consideration.

    ## **Use instead:**
    **No suggestions**.
    """

    def visit_AlterTableCmd(
        self,
        ancestors: visitors.Ancestor,
        node: ast.AlterTableCmd,
    ) -> None:
        """Visit AlterTableCmd."""
        if node.subtype in (
            enums.AlterTableType.AT_DropConstraint,
            enums.AlterTableType.AT_DropNotNull,
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Constraint `{node.name}` removal detected",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                ),
            )



"""Formatter for column."""

from pglast import ast, enums, stream, printers


@printers.node_printer(ast.ColumnDef, override=True)
def column_def(node: ast.ColumnDef, output: stream.RawStream) -> None:
    """Printer for ColumnDef."""
    if node.colname:
        output.print_name(node.colname)
        output.space()

    if node.typeName:
        output.print_name(node.typeName)
        if node.storage_name:
            output.write("STORAGE")
            output.space()
            output.write(node.storage_name)

        if node.compression:
            output.space()
            output.write("COMPRESSION")
            output.space()
            output.print_name(node.compression)
    elif node.constraints:
        output.write("WITH OPTIONS")
        output.space()

    if node.fdwoptions:
        output.space()
        output.write("OPTIONS")
        output.space()
        with output.expression(need_parens=True):
            output.print_list(node.fdwoptions, ",")

    if node.collClause:
        output.print_node(node.collClause)

    if node.constraints:
        # move NOT NULL constraint to the end
        constraints = [
            x for x in node.constraints if x.contype != enums.ConstrType.CONSTR_NOTNULL
        ] + [x for x in node.constraints if x.contype == enums.ConstrType.CONSTR_NOTNULL]
        output.print_list(constraints, "", standalone_items=False)



"""Checker for typed table."""

from pglast import ast, visitors

from pgrubic.core import linter


class TypedTable(linter.BaseChecker):
    """## **What it does**
    Checks for typed table.

    ## **Why not?**
    A typed table takes its structure from a composite type.

    A typed table is tied to its type; for example the table will be dropped if the type
    is dropped (with DROP TYPE ... CASCADE).

    This creates a tight coupling between the table and its type. Any operations to be
    done to the columns of the typed table would have to be done through the type.

    ## **When should you?**
    Almost Never.

    ## **Use instead:**
    A template table, copying the structure with **LIKE**.
    """

    def visit_CreateStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateStmt,
    ) -> None:
        """Visit CreateStmt."""
        if node.ofTypename:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="A typed table is tightly coupled to its type",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use a template table, copying the structure with **LIKE**",
                ),
            )



"""Checker for yoda conditions."""

from pglast import ast, enums, visitors

from pgrubic import Operators
from pgrubic.core import linter


class StringifiedNull(linter.BaseChecker):
    """## **What it does**
    Checks for stringified NULL.

    ## **Why not?**
    NULL is not a data value, but a marker for an absent value, and it should not be
    quoted.

    Putting NULL in quotes makes it a string value, which is not the same as a NULL value.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Use NULL without quotes.
    """

    is_auto_fixable: bool = True

    def visit_String(
        self,
        ancestors: visitors.Ancestor,
        node: ast.String,
    ) -> ast.A_Const:
        """Visit String that is not part of an A_Expr."""
        if node.sval.upper() == "NULL" and not ancestors.find_nearest(ast.A_Expr):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Quoted NULL detected",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Remove quotes from NULL",
                ),
            )

            return self._fix_string(node=node)

        return None

    def _fix_string(self, node: ast.String) -> ast.A_Const:
        """Fix string violation."""
        return ast.A_Const(
            isnull=True,
            val=ast.ValUnion(value=node),
        )

    def visit_A_Expr(
        self,
        ancestors: visitors.Ancestor,
        node: ast.A_Expr,
    ) -> ast.NullTest:
        """Visit A_Expr."""
        if (
            isinstance(node.lexpr, ast.A_Const)
            and isinstance(node.lexpr.val, ast.String)
            and node.lexpr.val.sval.upper() == "NULL"
        ) or (
            isinstance(node.rexpr, ast.A_Const)
            and isinstance(node.rexpr.val, ast.String)
            and node.rexpr.val.sval.upper() == "NULL"
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Quoted NULL detected",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Remove quotes from NULL",
                ),
            )

            return self._fix_expression(node=node)

        return None  # pragma: no cover

    def _fix_expression(self, node: ast.A_Expr) -> ast.A_Const:
        """Fix expression violation."""
        if not isinstance(node.rexpr, ast.A_Const) or (
            isinstance(node.rexpr, ast.A_Const)
            and isinstance(node.rexpr.val, ast.String)
            and node.rexpr.val.sval.upper() != "NULL"
        ):
            lexpr = node.rexpr
        else:
            lexpr = node.lexpr

        null_type = (
            enums.NullTestType.IS_NULL
            if node.name[-1].sval == Operators.EQ
            else enums.NullTestType.IS_NOT_NULL
        )

        return ast.NullTest(arg=lexpr, nulltesttype=null_type)



"""Checker for multi-column partitioning."""

from pglast import ast, visitors

from pgrubic.core import linter


class MultiColumnPartitioning(linter.BaseChecker):
    """## **What it does**
    Checks for partitioning with more than one column.

    ## **Why not?**
    One of the main reasons to use partitioning is the improved performance achieved by
    partition pruning. Pruning in a multi-column partitioned table can easily be
    sub-optimal, leading to scanning of unnecessary partitions.

    ## **When should you?**
    If you know what you are doing and have a good reason to do so. Just don't
    use multi-column partitioning without deep consideration and knowledge of its
    intrinsics. An example of what could go wrong is
    [why-isnt-postgres-multicolumn-partition-pruning-smarter-than-this](https://stackoverflow.com/questions/69662835/why-isnt-postgres-multicolumn-partition-pruning-smarter-than-this){:target="_blank"}

    ## **Use instead:**
    Sub-partitioning.
    """

    def visit_PartitionSpec(
        self,
        ancestors: visitors.Ancestor,
        node: ast.PartitionSpec,
    ) -> None:
        """Visit PartitionSpec."""
        if len(node.partParams) > 1:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer partitioning by one key",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use sub-partitioning",
                ),
            )



"""Caching of formatted files."""

import os
import typing
import hashlib
import pathlib
import tempfile

import msgpack

import pgrubic
from pgrubic import PACKAGE_NAME
from pgrubic.core import config

CACHE_FILE_NAME_LENGTH: typing.Final[int] = 20

CACHE_DIR_ENVIRONMENT_VARIABLE: typing.Final[str] = f"{PACKAGE_NAME.upper()}_CACHE_DIR"

DEFAULT_CACHE_DIR: typing.Final[str] = ".pgrubic_cache"


class FileData(typing.NamedTuple):
    """Representation of file data."""

    size: int
    last_modified_time: float
    hashed_content: str


class Cache:
    """Caching of formatted files."""

    def __init__(
        self,
        *,
        config: config.Config,
    ) -> None:
        """Initialize variables."""
        self.config = config
        if (
            os.getenv(CACHE_DIR_ENVIRONMENT_VARIABLE)
            and str(self.config.cache_dir) == DEFAULT_CACHE_DIR
        ):
            self.config.cache_dir = pathlib.Path(
                os.environ[CACHE_DIR_ENVIRONMENT_VARIABLE],
            )

        self.cache_dir = config.cache_dir.resolve() / pgrubic.__version__
        self.cache_file = (
            self.cache_dir
            / hashlib.sha256(b"formatter.cache").hexdigest()[:CACHE_FILE_NAME_LENGTH]
        )

    def _read(self) -> dict[str, FileData]:
        """Read the cache file if it exists."""
        if not self.cache_file.exists():
            return {}

        with self.cache_file.open("rb") as f:
            cache: dict[str, tuple[int, float, str]] = msgpack.unpack(f)
            return {
                k: FileData(size=v[0], last_modified_time=v[1], hashed_content=v[2])
                for k, v in cache.items()
            }

    def _hash_digest(self, source: pathlib.Path) -> str:
        """Return hash digest of the content of source and config.

        Parameters:
        ----------
        source: pathlib.Path
            Path to the source file.

        Returns:
        -------
        str
            Hash digest of the content of source and config.
        """
        hasher = hashlib.sha256()
        hasher.update(source.read_bytes())
        hasher.update(msgpack.packb(self.config.format.__repr__()))
        return hasher.hexdigest()

    def _get_file_data(self, source: pathlib.Path) -> FileData:
        """Return file data for source.

        Parameters:
        ----------
        source: pathlib.Path
            Path to the source file.

        Returns:
        -------
        FileData
            File data for source.
        """
        file_stat = source.stat()
        hashed_content = self._hash_digest(source)
        return FileData(
            size=file_stat.st_size,
            last_modified_time=file_stat.st_mtime,
            hashed_content=hashed_content,
        )

    def _need_to_be_formatted(self, source: pathlib.Path) -> bool:
        """Check if source needs to be formatted.

        Parameters:
        ----------
        source: pathlib.Path
            Path to the source file.

        Returns:
        -------
        bool
            True if source needs to be formatted, False otherwise.
        """
        cache = self._read()
        resolved_source = source.resolve()
        cached_version = cache.get(str(resolved_source))
        if not cached_version:
            return True

        source_stat = resolved_source.stat()

        if source_stat.st_size != cached_version.size:
            return True

        if source_stat.st_mtime != cached_version.last_modified_time:
            return True

        new_file_hash = self._hash_digest(resolved_source)
        return new_file_hash != cached_version.hashed_content

    def filter_sources(self, sources: set[pathlib.Path]) -> set[pathlib.Path]:
        """Return sources that need to be formatted.

        Parameters:
        ----------
        sources: set[pathlib.Path]
            Set of source files.

        Returns:
        -------
        set[pathlib.Path]
            Set of sources that need to be formatted.
        """
        sources_to_be_formatted: set[pathlib.Path] = set()
        for source in sources:
            if self._need_to_be_formatted(source):
                sources_to_be_formatted.add(source)

        return sources_to_be_formatted

    def write(self, sources: set[pathlib.Path]) -> None:
        """Generate the cache data for sources and write a new cache file.

        Parameters:
        ----------
        sources: set[pathlib.Path]
            Set of source files.

        Returns:
        -------
        None
        """
        file_data: dict[str, FileData] = {
            str(source.resolve()): self._get_file_data(source) for source in sources
        }

        self.cache_dir.mkdir(parents=True, exist_ok=True)

        with tempfile.NamedTemporaryFile(
            dir=str(self.cache_file.parent),
            delete=False,
        ) as tf:
            data: dict[str, tuple[int, float, str]] = {
                k: (v.size, v.last_modified_time, v.hashed_content)
                for k, v in file_data.items()
            }
            msgpack.pack(data, tf)

        pathlib.Path.replace(pathlib.Path(tf.name), self.cache_file)



"""Checker for missing replace in function."""

from pglast import ast, visitors

from pgrubic.core import linter


class MissingReplaceInFunction(linter.BaseChecker):
    """## **What it does**
    Checks for replace in function creation.

    ## **Why not?**
    `CREATE OR REPLACE FUNCTION` simplifies the process of modifying existing functions,
    as you don't need to manually drop and recreate them.

    If you drop and then recreate a function, the new function is not the same entity as
    the old; you will have to drop existing rules, views, triggers, etc. that refer to the
    old function. Use CREATE OR REPLACE FUNCTION to change a function definition without
    breaking objects that refer to the function. It also maintains data integrity and
    consistency.

    ## **When should you?**
    If you don't need to modify an existing function.

    ## **Use instead:**
    CREATE OR REPLACE FUNCTION.
    """

    is_auto_fixable: bool = True

    def visit_CreateFunctionStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateFunctionStmt,
    ) -> None:
        """Visit CreateFunctionStmt."""
        if not node.replace and not node.is_procedure:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer create or replace for function",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use create or replace",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.CreateFunctionStmt) -> None:
        """Fix violation."""
        node.replace = True



"""Checker for keywords used as identifiers."""

from pglast import keywords

from pgrubic.core import linter
from pgrubic.rules.naming import CheckIdentifier


class KeywordIdentifier(CheckIdentifier):
    """## **What it does**
    Checks for keywords used as identifiers.

    ## **Why not?**
    According to the standard, reserved key words are the only real key words; they are
    never allowed as identifiers. Non-reserved key words only have a special meaning in
    particular contexts and can be used as identifiers in other contexts.

    PostgreSQL won't allow reserved keywords as identifiers without double quotes.
    This means that if you use reserved keywords as identifiers, you have to always double
    quote them. That is annoying enough by hand and error-prone.

    Eventhough, non-reserved keywords can be used as identifiers in certain contexts,
    it can be confusing and ambiguious. Also, there is nothing stopping non-reserved
    keywords from becoming reserved keywords in the future. So, it is best to avoid them
    altogether.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Choose a name that is not a keyword.
    """

    def _check_identifier(
        self,
        identifier: str,
        line_number: int,
        column_offset: int,
        line: str,
        statement_location: int,
    ) -> None:
        """Check for keywords used as identifiers."""
        full_keywords: set[str] = (
            set(keywords.RESERVED_KEYWORDS)
            .union(
                set(keywords.UNRESERVED_KEYWORDS),
            )
            .union(keywords.COL_NAME_KEYWORDS)
            .union(keywords.TYPE_FUNC_NAME_KEYWORDS)
        )

        if identifier and identifier.strip().lower() in full_keywords:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=line_number,
                    column_offset=column_offset,
                    line=line,
                    statement_location=statement_location,
                    description=f"Keyword `{identifier}` used as an identifier",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Choose a name that is not a keyword",
                ),
            )



"""Checker for duplicate indexes."""

import typing

from pglast import ast, visitors

from pgrubic.core import linter


class DuplicateIndex(linter.BaseChecker):
    """## **What it does**
    Checks for duplicate indexes (exact match).

    ## **Why not?**
    Having duplicate indexes can negatively affect performance of database operations in
    several ways, some of which are as follows:

    - Slower Write Operations
    - Maintenance Overhead
    - Increased Storage Costs

    In summary, indexes are not cheap and what is worst is having them duplicated.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Remove the duplicate.
    """

    seen_indexes: typing.ClassVar[list[typing.Any]] = []

    def visit_IndexStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.IndexStmt,
    ) -> None:
        """Visit IndexStmt."""
        index_definition = (
            node.relation,
            node.indexParams,
            node.indexIncludingParams,
            node.whereClause,
        )

        if index_definition in self.seen_indexes:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Duplicate index detected",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Remove duplicate indexes",
                ),
            )

        self.seen_indexes.append(index_definition)



"""Checker for column drop."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class DropColumn(linter.BaseChecker):
    """## **What it does**
    Checks dropping of column.

    ## **Why not?**
    Not only that mistakenly dropping a column can cause data loss, applications that rely
    on the column will break.

    If any part of the application code, database procedures, views, or reports use
    the column, dropping it will cause errors and potentially disrupt business operations.

    Removing a column from a table may appear to be a reversible action, but it is not.
    Even when you can recover all the data in the column, you cannot restore the column in
    a way that makes the table look exactly as it did before.

    In postgres, **DROP COLUMN** form does not physically remove the column, but simply
    makes it invisible to SQL operations. Subsequent INSERT and UPDATE operations in the
    table will store a NULL value for the column.

    ## **When should you?**
    After updating clients that rely on the column to stop referencing the column and you
    really want to discard the data in the column.

    ## **Use instead:**
    You can either keep the column as nullable or drop it once it is no longer being
    referenced by clients.
    """

    def visit_AlterTableCmd(
        self,
        ancestors: visitors.Ancestor,
        node: ast.AlterTableCmd,
    ) -> None:
        """Visit AlterTableCmd."""
        if node.subtype == enums.AlterTableType.AT_DropColumn:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Drop column detected",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Keep the column as nullable"
                    " or drop it once it is no longer being used",
                ),
            )



"""Formatter for index."""

from pglast import ast, stream, printers

from pgrubic.formatters.ddl import IF_NOT_EXISTS


@printers.node_printer(ast.IndexStmt, override=True)
def index_stmt(node: ast.IndexStmt, output: stream.RawStream) -> None:
    """Printer for IndexStmt."""
    output.write("CREATE")
    output.space()

    if node.unique:
        output.write("UNIQUE")
        output.space()

    output.write("INDEX")

    if node.concurrent:
        output.space()
        output.write("CONCURRENTLY")

    if node.if_not_exists:
        output.space()
        output.write(IF_NOT_EXISTS)

    if node.idxname:
        output.space()
        output.print_name(node.idxname)

    output.newline()

    with output.push_indent(4):
        output.write("ON")
        output.space()
        output.print_node(node.relation)

        if node.accessMethod != "btree":
            output.write("USING")
            output.space()
            output.print_name(node.accessMethod)

        output.space()
        output.swrite("(")
        output.print_list(node.indexParams, standalone_items=False)
        output.swrite(")")

        if node.indexIncludingParams:
            output.space()
            output.write("INCLUDE")
            output.space()
            output.swrite("(")
            output.print_list(node.indexIncludingParams, standalone_items=False)
            output.swrite(")")

        if node.nulls_not_distinct:
            output.newline()
            output.indent(1)
            output.write("NULLS NOT DISTINCT")

        if node.options:
            output.newline()
            output.indent(2)
            output.write("WITH")
            output.space()
            with output.expression(need_parens=True):
                output.print_list(node.options, standalone_items=False)

        if node.tableSpace:
            output.newline()
            output.indent()
            output.write("TABLESPACE")
            output.space()
            output.print_name(node.tableSpace)

        if node.whereClause:
            output.newline()
            output.indent(1)
            output.write("WHERE")
            output.space()
            output.print_node(node.whereClause)



"""Checker for truncate table."""

from pglast import ast, visitors

from pgrubic.core import linter


class TruncateTable(linter.BaseChecker):
    """## **What it does**
    Checks truncating of table.

    ## **Why not?**
    Truncating a table can easily break applications that rely on the data in the table.

    If any part of the application code, database procedures, views, or reports use
    the data, truncating it will cause errors and potentially disrupt business operations.

    ## **When should you?**
    If the data in the table is no longer needed by clients, probably a test data.

    ## **Use instead:**
    No suggestions.
    """

    def visit_TruncateStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.TruncateStmt,
    ) -> None:
        """Visit TruncateStmt."""
        self.violations.add(
            linter.Violation(
                rule_code=self.code,
                rule_name=self.name,
                rule_category=self.category,
                line_number=self.line_number,
                column_offset=self.column_offset,
                line=self.line,
                statement_location=self.statement_location,
                description="Truncate table detected",
                is_auto_fixable=self.is_auto_fixable,
                is_fix_enabled=self.is_fix_enabled,
            ),
        )



"""pgrubic."""

import enum
import typing
import pathlib
import tomllib
import importlib.metadata

from pglast import ast

PACKAGE_NAME: typing.Final[str] = "pgrubic"

WORKERS_ENVIRONMENT_VARIABLE: typing.Final[str] = f"{PACKAGE_NAME.upper()}_WORKERS"

DEFAULT_WORKERS: typing.Final[int] = 4

DOCUMENTATION_URL: typing.Final[str] = "https://bolajiwahab.github.io/pgrubic"

PACKAGE_DIRECTORY: typing.Final[pathlib.Path] = pathlib.Path(__file__).resolve().parent

SOURCE_DIRECTORY: typing.Final[pathlib.Path] = PACKAGE_DIRECTORY.parent

PROJECT_DIRECTORY: typing.Final[pathlib.Path] = SOURCE_DIRECTORY.parent

RULES_DIRECTORY: typing.Final[pathlib.Path] = PACKAGE_DIRECTORY / "rules/"

FORMATTERS_DIRECTORY: typing.Final[pathlib.Path] = PACKAGE_DIRECTORY / "formatters/"

pyproject_file = pathlib.Path(PROJECT_DIRECTORY / "pyproject.toml")
if pyproject_file.exists():
    with pathlib.Path(PROJECT_DIRECTORY / "pyproject.toml").open("rb") as f:
        pyproject_config = tomllib.load(f)
        __version__: str = pyproject_config["project"]["version"]
else:
    __version__ = importlib.metadata.version(PACKAGE_NAME)  # pragma: no cover

RULES_BASE_MODULE: typing.Final[str] = f"{PACKAGE_NAME}/rules/"

FORMATTERS_BASE_MODULE: typing.Final[str] = f"{PACKAGE_NAME}/formatters/"


def get_fully_qualified_name(node: tuple[ast.Node]) -> str:
    """Get fully qualified type name.

    Parameters:
    ----------
    node: tuple[ast.Node]
        Node to get fully qualified name for.

    Returns:
    -------
    str
        Fully qualified type name.

    """
    if isinstance(node, ast.String):
        return str(node.sval)

    return ".".join(n.sval for n in node)


@enum.unique
class Operators(enum.StrEnum):
    """Operators."""

    EQ = "="
    NOT_EQ = "<>"



"""Formatter for rename."""

from pglast import ast, enums, stream, printers
from pglast.printers.ddl import OBJECT_NAMES


@printers.node_printer(ast.RenameStmt, override=True)
def rename_stmt(node: ast.RenameStmt, output: stream.RawStream) -> None:
    """Printer for RenameStmt."""
    objtype = node.renameType
    output.write("ALTER")
    output.space()
    if objtype == enums.ObjectType.OBJECT_TABCONSTRAINT:
        output.write("TABLE")
    elif objtype == enums.ObjectType.OBJECT_DOMCONSTRAINT:  # pragma: no cover
        output.write("DOMAIN")
    elif objtype == enums.ObjectType.OBJECT_ROLE:  # pragma: no cover
        output.write("ROLE")
    else:
        output.write(
            OBJECT_NAMES[
                (
                    node.relationType
                    if objtype
                    in (enums.ObjectType.OBJECT_ATTRIBUTE, enums.ObjectType.OBJECT_COLUMN)
                    else objtype
                )
            ],
        )
    output.space()

    if node.missing_ok:
        output.write("IF EXISTS")
        output.space()

    if objtype in (
        enums.ObjectType.OBJECT_SCHEMA,
        enums.ObjectType.OBJECT_DATABASE,
        enums.ObjectType.OBJECT_ROLE,
        enums.ObjectType.OBJECT_TABLESPACE,
    ):
        output.print_name(node.subname)
    elif objtype in (  # pragma: no cover
        enums.ObjectType.OBJECT_RULE,
        enums.ObjectType.OBJECT_POLICY,
        enums.ObjectType.OBJECT_TRIGGER,
    ):
        output.print_name(node.subname)
        output.space()
        output.write("ON")
        output.space()
        output.print_node(node.relation)
    elif node.relation:
        output.print_node(node.relation)
    elif objtype in (
        enums.ObjectType.OBJECT_OPFAMILY,
        enums.ObjectType.OBJECT_OPCLASS,
    ):  # pragma: no cover
        method, *name = node.object
        output.print_name(name)
        output.space()
        output.write("USING")
        output.space()
        output.print_symbol(method)
    else:
        output.print_name(node.object)
    output.newline()
    output.indent(4)
    output.write("RENAME")
    output.space()

    if objtype == enums.ObjectType.OBJECT_COLUMN:
        output.write("COLUMN")
        output.space()
        output.print_name(node.subname)
    elif objtype == enums.ObjectType.OBJECT_TABCONSTRAINT:
        output.write("CONSTRAINT")
        output.space()
        output.print_name(node.subname)
    elif objtype == enums.ObjectType.OBJECT_ATTRIBUTE:  # pragma: no cover
        output.write("ATTRIBUTE")
        output.space()
        output.print_name(node.subname)
    elif objtype == enums.ObjectType.OBJECT_DOMCONSTRAINT:  # pragma: no cover
        output.writes("CONSTRAINT")
        output.print_name(node.subname)

    output.swrite("TO")
    output.space()
    output.print_name(node.newname)

    if node.behavior == enums.DropBehavior.DROP_CASCADE:  # pragma: no cover
        output.space()
        output.write("CASCADE")



"""Checker for cluster."""

from pglast import ast, visitors

from pgrubic.core import linter


class Cluster(linter.BaseChecker):
    """## **What it does**
    Checks cluster.

    ## **Why not?**
    When a table is being clustered, an **ACCESS EXCLUSIVE** lock is acquired on it.
    This prevents any other database operations (both reads and writes) from operating on
    the table until the **CLUSTER** is finished.
    This will cause downtime if the table is concurrently being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    Have a look at pg_repack as an alternative.
    """

    def visit_ClusterStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ClusterStmt,
    ) -> None:
        """Visit ClusterStmt."""
        self.violations.add(
            linter.Violation(
                rule_code=self.code,
                rule_name=self.name,
                rule_category=self.category,
                line_number=self.line_number,
                column_offset=self.column_offset,
                line=self.line,
                statement_location=self.statement_location,
                description="Cluster found",
                is_auto_fixable=self.is_auto_fixable,
                is_fix_enabled=self.is_fix_enabled,
                help="Have a look at pg_repack as an alternative",
            ),
        )



"""Checker for serial types."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class Serial(linter.BaseChecker):
    """## **What it does**
    Checks for usage of serial types.

    ## **Why not?**
    The serial types have some weird behaviors that make schema, dependency, and
    permission management unnecessarily cumbersome.

    ## **When should you?**
    - If you need support to PostgreSQL older than version 10.
    - In certain combinations with table inheritance (but see there)
    - More generally, if you somehow use the same sequence for multiple tables, although
      in those cases an explicit declaration might be preferable over the serial types.

    ## **Use instead:**
    For new applications, identity columns should be used.
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        # While serial is not a postgres type, it can be used as a shortcut to type a
        # column as either smallint, int or bigint with a backing sequence.
        # Using serial in column definition is fine and works but it does not work
        # in cases of trying to change a column type since serial is not a type e.g
        # `ALTER TABLE table_name ALTER COLUMN column_name TYPE serial;`
        # results in error `ERROR:  type "serial" does not exist`
        # But such cases are still parseable. For this reason, we skip such statements
        alter_table_cmd: visitors.Ancestor = ancestors.find_nearest(ast.AlterTableCmd)

        if (
            (
                alter_table_cmd
                and alter_table_cmd.node.subtype == enums.AlterTableType.AT_AddColumn
            )
            or ancestors.find_nearest(ast.CreateStmt)
        ) and node.typeName.names[-1].sval in ["smallserial", "serial", "bigserial"]:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer identity column over serial types",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use identity column",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "bigint",
                },
            ),
        )

        node.constraints = (
            *(node.constraints or []),
            ast.Constraint(
                contype=enums.ConstrType.CONSTR_IDENTITY,
                generated_when=enums.ATTRIBUTE_IDENTITY_ALWAYS,
            ),
        )



"""Rules for naming."""

import abc

from pglast import ast, visitors

from pgrubic.core import linter


class ABCBaseCheckerMeta(abc.ABCMeta, linter.CheckerMeta):
    """Combine ABCMeta and CheckerMeta."""


class CheckIdentifier(abc.ABC, linter.BaseChecker, metaclass=ABCBaseCheckerMeta):
    """Check identifier."""

    @abc.abstractmethod
    def _check_identifier(
        self,
        *,
        identifier: str,
        line_number: int,
        column_offset: int,
        line: str,
        statement_location: int,
    ) -> None:
        """Check identifier for violations."""

    def visit_CreateStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateStmt,
    ) -> None:
        """Visit CreateStmt."""
        self._check_identifier(
            identifier=node.relation.relname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        self._check_identifier(
            identifier=node.colname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_ViewStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ViewStmt,
    ) -> None:
        """Visit ViewStmt."""
        self._check_identifier(
            identifier=node.view.relname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_IndexStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.IndexStmt,
    ) -> None:
        """Visit IndexStmt."""
        self._check_identifier(
            identifier=node.idxname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_CreateSeqStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateSeqStmt,
    ) -> None:
        """Visit CreateSeqStmt."""
        self._check_identifier(
            identifier=node.sequence.relname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_CreateSchemaStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateSchemaStmt,
    ) -> None:
        """Visit CreateSchemaStmt."""
        self._check_identifier(
            identifier=node.schemaname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_CreateFunctionStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateFunctionStmt,
    ) -> None:
        """Visit CreateFunctionStmt."""
        self._check_identifier(
            identifier=node.funcname[-1].sval,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if node.conname is not None:
            self._check_identifier(
                identifier=node.conname,
                line_number=self.line_number,
                column_offset=self.column_offset,
                line=self.line,
                statement_location=self.statement_location,
            )

    def visit_CreatedbStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreatedbStmt,
    ) -> None:
        """Visit CreatedbStmt."""
        self._check_identifier(
            identifier=node.dbname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_CreateRoleStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateRoleStmt,
    ) -> None:
        """Visit CreateRoleStmt."""
        self._check_identifier(
            identifier=node.role,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_CreateTableSpaceStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateTableSpaceStmt,
    ) -> None:
        """Visit CreateTableSpaceStmt."""
        self._check_identifier(
            identifier=node.tablespacename,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_CreateTrigStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateTrigStmt,
    ) -> None:
        """Visit CreateTrigStmt."""
        self._check_identifier(
            identifier=node.trigname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_CreateEnumStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateEnumStmt,
    ) -> None:
        """Visit CreateEnumStmt."""
        self._check_identifier(
            identifier=node.typeName[-1].sval,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_RuleStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.RuleStmt,
    ) -> None:
        """Visit RuleStmt."""
        self._check_identifier(
            identifier=node.rulename,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_IntoClause(
        self,
        ancestors: visitors.Ancestor,
        node: ast.IntoClause,
    ) -> None:
        """Visit IntoClause."""
        self._check_identifier(
            identifier=node.rel.relname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_CompositeTypeStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CompositeTypeStmt,
    ) -> None:
        """Visit IntoClause."""
        self._check_identifier(
            identifier=node.typevar.relname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )

    def visit_RenameStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.RenameStmt,
    ) -> None:
        """Visit RenameStmt."""
        self._check_identifier(
            identifier=node.newname,
            line_number=self.line_number,
            column_offset=self.column_offset,
            line=self.line,
            statement_location=self.statement_location,
        )
